/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.rpc.Stub;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.MessageContext;
import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.PositiveInteger;

import org.globus.axis.transport.local.LocalTransportUtils;
import org.globus.exec.generated.InternalStateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.ManagedExecutableJobResourcePropertiesType;
import org.globus.exec.generated.ManagedExecutableJobResourceStateType;
import org.globus.exec.generated.MultiJobDescriptionType;
import org.globus.exec.generated.ServiceLevelAgreementType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.monitoring.JobStateChangeListener;
import org.globus.exec.service.exec.utils.LocalTransportClient;
import org.globus.exec.service.factory.FactoryServiceConfiguration;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.job.ManagedJobHome;
import org.globus.exec.service.utils.AuditDatabaseConfiguration;
import org.globus.exec.service.utils.ConfigurationHelper;
import org.globus.exec.service.utils.GlobusShToolsProperties;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.audit.AuditDatabaseAppender;
import org.globus.exec.utils.service.ManagedJobHelper;
import org.globus.util.I18n;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.PersistentResource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.container.ServiceContainer;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.PerformanceLog;
import org.globus.wsrf.utils.XmlPersistenceHelper;


public class ManagedExecutableJobHome extends       ResourceHomeImpl
                                      implements    JobStateChangeListener,
                                                    ManagedJobHome
{

    private static final Log logger =
        LogFactory.getLog(ManagedExecutableJobHome.class.getName());

    static I18n i18n = I18n.getI18n(
        org.globus.exec.utils.Resources.class.getName());

    private static ManagedExecutableJobHome INSTANCE = null;

    private static Map userHomeMap
        = Collections.synchronizedMap(new HashMap());

    private static AuditDatabaseConfiguration
        AUDIT_DATABASE_CONFIGURATION = null;

    private static PerformanceLog performanceLogger = new PerformanceLog(
        ManagedExecutableJobHome.class.getName() + ".performance");

    private boolean recovered = false;
    private boolean initialized = false;

    private MessageContext cachedContext = null;
    private Hashtable idempotenceIdMap = null;

    static
    {
        LocalTransportUtils.init();
    }

    public ManagedExecutableJobHome() {
        //DO NOT DO ANYTHING HERE. PUT IT IN initialize().
        logger.debug("Entering ManagedExecutableJobHome()");
        logger.debug("Leaving ManagedExecutableJobHome()");
    }

    public synchronized void initialize()
        throws                              Exception
    {
        if (this.initialized)
        {
            return;
        }
        logger.debug("Entering initialize()");

        /*cache the RunQueue instance now so that we don't incur the class
          loading penalty during the first job submission*/
        RunQueue.getInstance();

        this.idempotenceIdMap = new Hashtable();

        this.cachedContext = MessageContext.getCurrentContext();

        //Key deserialization configuration (easier here than in JNDI config)
        this.keyTypeName = ManagedJobConstants.RESOURCE_KEY_QNAME;

        super.initialize();

        StateMachine.init();

        AuditDatabaseConfiguration auditDatabaseConfiguration
            = getAuditDatabaseConfiguration();
        AuditDatabaseAppender.initialize(
            auditDatabaseConfiguration.getDriverClass(),
            auditDatabaseConfiguration.getUrl(),
            auditDatabaseConfiguration.getUser(),
            auditDatabaseConfiguration.getPassword());

        this.initialized = true;
        notifyAll();

        logger.debug("Leaving initialize()");
    }

    public boolean isInitialized()
    {
        return this.initialized;
    }

    /** This method assures that any jobs which were being processed by the
     *  state machine can continue to do so after container recovery.
     *  Subscriptions to the JobStateMonitor are also restored.
     *  It is intended to be called by ManagedJobFactoryResource after starting
     *  the JobStateMonitor.
     * <p>
     * <b>Precondition: </b>isResourcePersistent()
     */
    public synchronized void recover() throws ResourceException
    {
        if (this.recovered)
        {
            return;
        }
        logger.debug("Entering recover()");

        if (!isResourcePersistent()) {
            String errorMessage = i18n.getMessage(
                Resources.PRECONDITION_VIOLATION,
                "!this.isResourcePersistent()");
            throw new RuntimeException(errorMessage);
        }
        logger.debug("Loading the home's list of resource keys");

        List resourceKeyValueList = null;
        try {
            XmlPersistenceHelper persistenceHelper = new XmlPersistenceHelper(
                ManagedExecutableJobResourceStateType.class);
            resourceKeyValueList = persistenceHelper.list();
        } catch (Exception e) {
            throw new ResourceException(
                i18n.getMessage("persistedResourcePropertiesDirectoryMissing"),
                e);
        }

        Iterator it = resourceKeyValueList.iterator();
        while (it.hasNext()) {
            String keyValue = (String) it.next();
            ResourceKey key = ManagedJobHelper.getResourceKey(keyValue);
            ManagedExecutableJobResource resource
                = (ManagedExecutableJobResource) find(key);
            //this.add(key, null);
            logger.info("Recovered resource with ID " + keyValue + ".");

            //Restart resource state processing
            synchronized (resource.getStateTransitionMutex())
            {
                ManagedExecutableJobResourceDataType resourceData
                    = (ManagedExecutableJobResourceDataType)
                        resource.getResourceData();
                InternalStateEnumeration restartInternalState =
                    resourceData.getRestartInternalState();
                InternalStateEnumeration internalState =
                    resourceData.getInternalState();
                if (   (restartInternalState == null)
                    || !internalState.equals(InternalStateEnumeration.Restart))
                {
                    resourceData.setRestartInternalState(internalState);

                    resourceData.setInternalState(
                        InternalStateEnumeration.Restart);
                }

                if (logger.isDebugEnabled()) {
                    logger.debug("Adding to RunQueue.");
                }
                RunQueue.getInstance().add(
                    (ResourceKey)resource.getID());
            }
        }

        this.recovered = true;

        logger.debug("Leaving recover()");
    }

    /**
     * If the messageID is not null, it is used as the resource ID.
     * If the messageID is null, the key value is computed automatically as a UUID.
     * @throws Exception
     * @return EndpointReferenceType
     */
    public ResourceKey create(
        Calendar                            initialTerminationTime,
        Calendar                            currentTime,
        String                              localResourceManager,
        JobDescriptionType                  jobDescription,
        String                              userSubject,
        String                              localUserId,
        String                              idempotenceId,
        String                              serviceURL)
        throws                              Exception
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("Entering create()");
        }

        SimpleResourceKey key = null;
        ManagedExecutableJobResource resource = null;

        if (idempotenceId != null)
        {
            key = (SimpleResourceKey) this.idempotenceIdMap.get(idempotenceId);

            if (key != null)
            {
                try
                {
                    resource = (ManagedExecutableJobResource) find(key);
                    return key;
                }
                catch (org.globus.wsrf.NoSuchResourceException nsre)
                {
                    /* not a duplicate idempotenceId */
                    logger.debug(
                        "JobID Specified, but not a duplicate job id");
                }
            }
        }

        if (jobDescription instanceof MultiJobDescriptionType)
        {
            throw new RemoteException(i18n.getMessage(
                "multiJobNotSupported",
                new Object[]{localResourceManager}));
        }

        //cache the notification consumer home while we have a context
        ManagedExecutableJobHome.getNotificationConsumerManager();

        key = ManagedJobHelper.getResourceKey(uuidGen.nextUUID());
        if (idempotenceId != null)
        {
            this.idempotenceIdMap.put(idempotenceId, key);
        }

        EndpointReferenceType epr = AddressingUtils.createEndpointReference(
            serviceURL, key);
        String handle = ManagedJobHelper.getHandle(epr);

        try {

            if (logger.isTraceEnabled())
            {
                logger.trace("at find()");
            }
            resource = (ManagedExecutableJobResource) find(key);
        }
        catch (NoSuchResourceException nsre) {
            logger.debug("new resource being created");
        }

        if (resource == null) {
            //create new resource object
            if (logger.isTraceEnabled())
            {
                logger.trace("at createNewInstance()");
            }
            resource = (ManagedExecutableJobResource)
                this.createNewInstance();
        }
        else
        {
            //repeat create request, so just return the resource key
            return key;
        }

        //initialize resource
        /*TODO: percolate initializeWholeState() functionality
                to ReflectionResource*/
        //ManagedExecutableJobResource is really a
        //"ManagedJobResourceImpl"
        if (logger.isTraceEnabled())
        {
            logger.trace("at initializeWholeState()");
        }
        resource.initializeWholeState(
            createResourceState(
                initialTerminationTime,
                currentTime,
                localResourceManager,
                jobDescription,
                userSubject,
                localUserId,
                handle,
                idempotenceId),
            ManagedJobConstants.RESOURCE_PROPERTY_SET,
            key);
        logger.debug(
            "Created ManagedExecutableJobResource with key value "
            + key.getValue());

        //Very important: store the persistent resource initially, because
        //the garbage collecting of the object will not trigger it.
        if (isResourcePersistent()) {
            ((PersistentResource)resource).store();
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("at add()");
        }
        this.add(key, resource);

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving create()");
        }

        return key;
    }

    private static NotificationConsumerManager NOTIFICATION_CONSUMER_MANAGER
        = null;;

    public static NotificationConsumerManager getNotificationConsumerManager()
    {
        if (NOTIFICATION_CONSUMER_MANAGER == null)
        {
            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String wsTransportProtocol
                = factoryHomeConfiguration.getStagingProtocol();
            if (wsTransportProtocol == null)
            {
                wsTransportProtocol = ServiceHost.getDefaultProtocol();
            }
            if (wsTransportProtocol.equals("http"))
            {
                NOTIFICATION_CONSUMER_MANAGER
                    = NotificationConsumerManager.getInstance();
            } else
            {
                Map properties = new HashMap();
                properties.put(ServiceContainer.CLASS,
                    "org.globus.wsrf.container.GSIServiceContainer");
                NOTIFICATION_CONSUMER_MANAGER
                    = NotificationConsumerManager.getInstance(properties);
            }
        }

        return NOTIFICATION_CONSUMER_MANAGER;
    }

    public boolean isResourcePersistent() {
        return PersistentResource.class.isAssignableFrom(this.resourceClass);
    }

    /**
     * Creates entire initial state of the resource, including internal
     * (non-published) data.
     * @return ManagedExecutableJobResourceStateType state object,
     *         serializable to XML by Axis
     */
    private static ManagedExecutableJobResourceStateType createResourceState(
        Calendar                            initialTerminationTime,
        Calendar                            currentTime,
        String                              localResourceManager,
        JobDescriptionType                  jobDescription,
        String                              userSubject,
        String                              localUserId,
        String                              handle,
        String                              idempotenceId)
    {
        ManagedExecutableJobResourceStateType resourceState =
            new ManagedExecutableJobResourceStateType();

        //initializing resource properties:
        ManagedExecutableJobResourcePropertiesType resourceBean =
            createResourcePropertiesBean(
                initialTerminationTime,
                currentTime,
                jobDescription,
                userSubject,
                localUserId);
        resourceState.setResourceProperties(resourceBean);

        ManagedExecutableJobResourceDataType internalData =
            new ManagedExecutableJobResourceDataType();

        //================= managed job: implementation data
        internalData.setInternalState(InternalStateEnumeration.None);
        internalData.setStarted(false);
        internalData.setHoldState(jobDescription.getHoldState());

        internalData.setLocalResourceManager(localResourceManager);
        internalData.setHandle(handle);

        //usage data
        internalData.setCreationTime(new Date());

        //audit data
        internalData.setIdempotenceId(idempotenceId);

        //================= rendezvous: implementation data
        //nothing (at creation time, values decided by impl of rendezvous)

        resourceState.setResourceData(internalData);

        return resourceState;
    }

    private static ManagedExecutableJobResourcePropertiesType
        createResourcePropertiesBean(
            Calendar                        initialTerminationTime,
            Calendar                        currentTime,
            JobDescriptionType              jobDescription,
            String                          userSubject,
            String                          localUserId)
    {
        ManagedExecutableJobResourcePropertiesType jobResourceProperties =
            new ManagedExecutableJobResourcePropertiesType();

        //================= scheduled termination: resource properties
        jobResourceProperties.setTerminationTime(initialTerminationTime);
        jobResourceProperties.setCurrentTime(currentTime);

        //================= managed job: resource properties
        jobResourceProperties.setState(StateEnumeration.Unsubmitted);
        PositiveInteger countObj = jobDescription.getCount();
        if (countObj == null) { //number of processes
            countObj = DEFAULT_PROCESS_COUNT;
            jobDescription.setCount(countObj);
        }
        jobResourceProperties.setServiceLevelAgreement(
            new ServiceLevelAgreementType(jobDescription, null));//job, multiJob
        jobResourceProperties.setLocalUserId(localUserId);
        jobResourceProperties.setUserSubject(userSubject);

        //================= rendezvous: resource properties
        jobResourceProperties.setCapacity(countObj.intValue());
        jobResourceProperties.setRegistrantData(null);
        jobResourceProperties.setRendezvousCompleted(false);

        return jobResourceProperties;
    }

    private static final PositiveInteger DEFAULT_PROCESS_COUNT =
        new PositiveInteger("1");

    private static UUIDGen uuidGen = UUIDGenFactory.getUUIDGen();

    public void jobStateChanged(
            ResourceKey                     resourceKey,
            String                          localJobId,
            Date                            timestamp,
            StateEnumeration                state,
            int                             exitCode)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Receiving jobStateChange event for resource key " +
                         resourceKey + " with: " +
                         "\ntimestamp\t" + timestamp +
                         "\n(new) state\t" + state.getValue() +
                         "\nexitCode\t" + exitCode);
        }
        ManagedExecutableJobResource resource;
        try {
            resource = (ManagedExecutableJobResource) find(resourceKey);
        } catch (ResourceException re) {
            String errorMessage = i18n.getMessage(
                Resources.NOTIFICATION_TARGET_RESOURCE_NOT_FOUND);
            logger.error(errorMessage, re);
            return;
        }

        StateMachine.localJobStateChanged(
            resource, localJobId, timestamp, state, exitCode);
    }

    public static ManagedExecutableJobHome getHome()
        throws                              ResourceException
    {
        if (ManagedExecutableJobHome.INSTANCE == null)
        {
            try
            {
                FactoryServiceConfiguration factoryConfiguration =
                    (FactoryServiceConfiguration)ConfigurationHelper.
                    getServiceConfiguration();

                ManagedExecutableJobHome.INSTANCE
                    = (ManagedExecutableJobHome) ConfigurationHelper
                    .getResourceHome(
                    factoryConfiguration.getManagedExecutableJobServicePath());

                logger.debug("created MEJH instance");

            }
            catch(javax.naming.NamingException ne)
            {
                throw new ResourceException(ne);
            }
            catch(org.globus.wsrf.ResourceContextException rce)
            {
                throw new ResourceException(rce);
            }
        }

        return INSTANCE;
    }

    public static String getUserHomeDirectory(
        String                              localUserId)
    {
        String userHome = null;
        if (userHomeMap.containsKey(localUserId))
        {
            userHome = (String) userHomeMap.get(localUserId);
        }
        else
        {
            GlobusShToolsProperties toolPathProps
                = GlobusShToolsProperties.getInstance();
            String perl = toolPathProps.getProperty("GLOBUS_SH_PERL");
            String[] command = new String[] {
                perl,
                "-e",
                "print '' . (getpwnam('" +
                    localUserId + "'))[7] . \"\\n\""
            };

            if (logger.isDebugEnabled()) {
                StringBuffer commandBuffer = new StringBuffer(command.length*2);
                for (int index=0; index<command.length; index++) {
                    if (index > 0) {
                        commandBuffer.append(" ");
                    }
                    commandBuffer.append(command[index]);
                }
                logger.debug("Executing command:\n" + commandBuffer.toString());
            }

            BufferedReader homeReader = null;
            InputStream err = null;
            OutputStream out = null;
            try {
                Process proc = Runtime.getRuntime().exec(command);

                err = proc.getErrorStream();
                out = proc.getOutputStream();
                homeReader = new BufferedReader(
                    new InputStreamReader(proc.getInputStream()));

                userHome = homeReader.readLine();
                proc.waitFor();
            } catch (InterruptedException ie) {
            } catch (IOException ioe) {
            } finally {
                if (homeReader != null) {
                    try { homeReader.close(); } catch (IOException e) {}
                }
                if (out != null) {
                    try { out.close(); } catch (IOException e) {}
                }
                if (err != null) {
                    try { err.close(); } catch (IOException e) {}
                }
            }

            userHomeMap.put(localUserId, userHome);
        }

        return userHome;
    }

    public int getResourceCount()
    {
        return this.resources.size();
    }

    ManagedExecutableJobResource unsyncFind(ResourceKey key)
        throws                              ResourceException
    {
        ManagedExecutableJobResource resource
            = (ManagedExecutableJobResource)this.resources.get(key);
        if (resource == null) {
	    resource = (ManagedExecutableJobResource)find(key);
        }

        return resource;
    }

    public void enableLocalTransport(
        Stub                                stub)
    {
        LocalTransportClient.enableLocalTransport(
            this.cachedContext,
            stub);
    }

    public static AuditDatabaseConfiguration getAuditDatabaseConfiguration()
    {
        if (AUDIT_DATABASE_CONFIGURATION == null)
        {
            try
            {
                AUDIT_DATABASE_CONFIGURATION
                    = (AuditDatabaseConfiguration) ConfigurationHelper
                    .getAuditDatabaseConfiguration();
            }
            catch (Exception e)
            {
                throw new RuntimeException(e);
            }
        }

        return AUDIT_DATABASE_CONFIGURATION;
    }

}
