/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.beans.IntrospectionException;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.lang.reflect.InvocationTargetException;

import java.util.Hashtable;
import java.util.Enumeration;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.generated.ExtensionsType;
import org.globus.exec.generated.FilePairType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.NameValuePairType;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.exec.utils.JobDescriptionMap;

import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;

import org.globus.wsrf.utils.AnyHelper;

public class PerlJobDescription extends JobDescriptionMap {
    static final String RSL_ARGUMENTS = "arguments";
    static final String RSL_COUNT = "count";
    static final String RSL_DIRECTORY = "directory";
    static final String RSL_DRY_RUN = "dryrun";
    static final String RSL_ENVIRONMENT = "environment";
    static final String RSL_GRAM_MY_JOB = "grammyjobtype";
    static final String RSL_HOST_COUNT = "hostcount";
    static final String RSL_JOB_TYPE = "jobtype";
    static final String RSL_MAX_CPU_TIME = "maxcputime";
    static final String RSL_MAX_MEMORY = "maxmemory";
    static final String RSL_MAX_TIME = "maxtime";
    static final String RSL_MAX_WALL_TIME = "maxwalltime";
    static final String RSL_MIN_MEMORY = "minmemory";
    static final String RSL_REMOTE_IO_URL = "remoteiourl";
    static final String RSL_SAVE_STATE = "savestate";
    static final String RSL_TWO_PHASE = "twophase";
    static final String RSL_FILE_STAGE_IN = "filestagein";
    static final String RSL_FILE_STAGE_OUT = "filestageout";
    static final String RSL_STDOUT = "stdout";
    static final String RSL_STDERR = "stderr";
    static final String RSL_STDIN = "stdin";
    static final String RSL_FILE_CLEAN_UP = "filecleanup";
    static final String RSL_LIBRARY_PATH = "librarypath";

    static Log logger = LogFactory.getLog(PerlJobDescription.class);

    public PerlJobDescription() { }

    public PerlJobDescription(
        JobDescriptionType                  jobDescription)
        throws                              IntrospectionException,
                                            IllegalAccessException,
                                            InvocationTargetException
    {
        super(jobDescription);
    }

    public PerlJobDescription(
        JobDescriptionType                  jobDescription,
        NameValuePairType[]                 extraEnvironment)
        throws                              IntrospectionException,
                                            IllegalAccessException,
                                            InvocationTargetException
    {
        super(jobDescription);

        //convert "argument" to "arguments"
        String[] arguments = (String[]) remove("argument");
        put("arguments", arguments);

        if (extraEnvironment == null) {
            return;
        }

        //get client-specified environment array
        NameValuePairType[] environment
            = (NameValuePairType[]) get(RSL_ENVIRONMENT);
        if (environment == null) {
            environment = new NameValuePairType[0];
        }

        //create new environment array
        NameValuePairType[] extendedEnvironment
            = new NameValuePairType[environment.length+extraEnvironment.length];

        //concat old environment and extra environment
        System.arraycopy(
            environment, 0,
            extendedEnvironment, 0,
            environment.length);
        System.arraycopy(
            extraEnvironment, 0,
            extendedEnvironment, environment.length,
            extraEnvironment.length);

        //save extended environment array to job description map object
        put(RSL_ENVIRONMENT, extendedEnvironment);
    }

    public String toPerlString() {
        StringBuffer description = new StringBuffer();

        description.append("$description = {\n");

        java.util.Iterator attributeNames = keySet().iterator();

        while (attributeNames.hasNext()) {
            String origName = (String) attributeNames.next();
            String name = canonicalize(origName);

            if (logger.isDebugEnabled()) {
                logger.debug("processing attribute name "
                            + origName + "/" + name);
            }

            //staging is done by RFT
            if (   name.equals(RSL_FILE_STAGE_IN) )
            {
                // But GGW still needs the info
                TransferRequestType xferReq =  ((TransferRequestType)get(origName));

                if( xferReq == null )
                        continue;

                TransferType [] xfertype = xferReq.getTransfer();

                description.append("    filestagein => [ ");

                for(int i =0 ; i < xfertype.length; i++)
                {
                        String srcURL = xfertype[i].getSourceUrl();
                        String destURL= xfertype[i].getDestinationUrl();
                        description.append("'" + srcURL + " " + destURL + "',");
                }

                description.deleteCharAt(description.length()-1);
                description.append("],\n");
                continue;
            }

            //staging is done by RFT
            if (   name.equals(RSL_FILE_STAGE_OUT) )
            {
                // But GGW still needs the info
                TransferRequestType xferReq =  ((TransferRequestType)get(origName));

                if( xferReq == null )
                        continue;

                TransferType [] xfertype = xferReq.getTransfer();

                description.append("    filestageout => [");

                for(int i =0 ; i < xfertype.length; i++)
                {
                        String srcURL  = xfertype[i].getSourceUrl();
                        String destURL = xfertype[i].getDestinationUrl();
                        description.append("'" + srcURL + " " + destURL + "',");
                }
                description.deleteCharAt(description.length()-1);
                description.append("],\n");
                continue;

            }

            Object value = get(origName);

            //TODO move that code to set defaults directly in the SLA RP of the Job ?
            //Or have client just rely on specific GrdiFTP-enabled RPs?
            if (value == null) {
                if (name.equals(RSL_DIRECTORY)) {
                    //TODO this will have to change for the multi-user version
                    description.append("    directory => [ '"
                            + System.getProperty("user.home") + "' ],\n");
                } else
                if (name.equals(RSL_STDIN)) {
                    logger.debug("appending stdin = /dev/null");
                    description.append(
                        "    stdin => [ '/dev/null' ],\n");
                } else
                if (name.equals(RSL_COUNT)) {
                    description.append("    count => [ 1 ],\n");
                } else
                if (name.equals(RSL_JOB_TYPE)) {
                    description.append(
                        "    jobtype => [ 'multiple' ],\n");
                }

                continue;
            }

            description.append("    ");

            Class valueClass = value.getClass();
            String valueClassName = valueClass.getName();

            if (logger.isDebugEnabled()) {
                logger.debug("Attribute " + name
                            + " is of type " + valueClassName);
            }

            if (valueClass.isArray()) {
                Class componentClass
                    = valueClass.getComponentType();
                if (logger.isDebugEnabled()) {
                    logger.debug("Attribute " + name
                                + " has component type "
                                + componentClass.getName());
                }
                if (componentClass.equals(NameValuePairType.class)) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(
                            "Attribute " + name + " being processed as "
                            + "NameValuePairType[]");
                    }
                    description.append(name).append(" => [ ").
                    append(makeStringFromNameValuePairs(
                        (NameValuePairType[]) value));
                    description.append(" ]");
                } else if (componentClass.equals(FilePairType.class)) {
                    description.append(name).append(" => [ ").
                    append(makeStringFromFilePairs((FilePairType[]) value)).
                    append(" ]");
                } else {
                    description.append(name).append(" => [ ").
                    append(makeStringList(value)).
                    append(" ]");
                }
            } else if (value instanceof Boolean) {
                description.append(name).append(" => [ '").
                append(((Boolean)value).booleanValue()?"yes":"no").
                append("' ]");
            } else if (value instanceof NameValuePairType) {
                NameValuePairType nvp = (NameValuePairType) value;
                description.append(name).append(" => [ [ ").
                append(makeString(nvp.getName())).
                append(", ").
                append(makeString(nvp.getValue())).
                append("]  ]");
            } else if (value instanceof FilePairType) {
                description.append(name).append(" => [ [ ").
                append(makeString(((FilePairType)value).getSourceFile())).
                append(", ").
                append(makeString(((FilePairType)value).getDestinationFile())).
                append("]  ]");
            } else {
                description.append(name).
                append(" => [ ").append(makeString(value)).append(" ]");
            }


            if (attributeNames.hasNext()) {
                description.append(",\n");
            } else {
                description.append("\n");
            }
        }

        description.append("};\n");

        return description.toString();
    }

    private String makeStringFromNestedList(Object o) {
        StringBuffer buffer = new StringBuffer();

        if (o instanceof Object[]) {
            Object[] tmp = (Object[]) o;
            buffer.append("[ ");
            for (int i = 0; i < tmp.length; i++) {
                buffer.append(makeStringFromNestedList(tmp[i]));

                if(i + 1 < tmp.length) {
                    buffer.append(", ");
                }
            }
            buffer.append(" ]");
        } else if (o instanceof String) {
            String tmp = (String) o;

            buffer.append(makeString(tmp));
        }

        return buffer.toString();
    }

    private String makeString(Object o) {
        char [] chars = null;
        if (o instanceof ExtensionsType)
        {
            ExtensionsType extensions = (ExtensionsType) o;
            try {
                chars = AnyHelper.toSingleString(
                    extensions.get_any()).toCharArray();
            } catch (Exception e)
            {
                logger.error("unable to convert extensions to string");
                return null;
            }
        }
        else
        {
            chars = o.toString().toCharArray();
        }
        StringBuffer str = new StringBuffer();
        int j = 0;

        str.append('\'');
        for (int i = 0; i < chars.length; i++) {
            if (chars[i] == '\''  || chars[i] == '\\') {
                str.append('\\');
                str.append(chars[i]);
            } else {
                str.append(chars[i]);
            }
        }

        str.append('\'');

        return str.toString();
    }

    private String makeStringList(Object values) {
        Class valuesClass = values.getClass();
        if (!valuesClass.isArray()) {
            return null;
        }

        return makeStringList((Object[]) values);
    }

    private String makeStringList(Object [] values) {
        StringBuffer buffer = new StringBuffer();

        for (int i = 0; i < values.length; i++) {
            buffer.append(makeString(values[i].toString()));

            if ((i + 1) < values.length) {
                buffer.append(", ");
            }
        }
        return buffer.toString();
    }

    private String makeStringFromNameValuePairs(NameValuePairType[] nvpairs) {
        StringBuffer buffer = new StringBuffer();

        for (int index=0; index<nvpairs.length; index++) {
            buffer.append("[ ").
            append(makeString(nvpairs[index].getName())).append(", ").
            append(makeString(nvpairs[index].getValue())).
            append(" ]");

            if ((nvpairs.length - index) > 0) {
                buffer.append(", ");
            }
        }

        return buffer.toString();
    }

    private String makeStringFromHash(Hashtable table) {
        StringBuffer buffer = new StringBuffer();
        Enumeration e = table.keys();

        while (e.hasMoreElements()) {
            String key = (String) e.nextElement();

            buffer.append("[ " + makeString(key) + ", "
                    + makeString((String) table.get(key)) + " ]");

            if (e.hasMoreElements()) {
                buffer.append(", ");
            }
        }
        return buffer.toString();
    }

    private String makeStringFromFilePairs(FilePairType[] filePairs) {
        StringBuffer buffer = new StringBuffer();

        for (int index=0; index<filePairs.length; index++) {
            buffer.append("[ ").
            append(makeString(filePairs[index].getSourceFile())).
            append(", ").
            append(makeString(filePairs[index].getDestinationFile())).
            append(" ]");

            if ((filePairs.length - index) > 0) {
                buffer.append(", ");
            }
        }

        return buffer.toString();
    }

    private String canonicalize(String s) {
        char [] chars = s.toCharArray();
        char [] canonical = new char[chars.length];
        int j = 0;

        for (int i = 0; i < chars.length; i++) {
            if (chars[i] == '_') {
                continue;
            } else {
                canonical[j++] = Character.toLowerCase(chars[i]);
            }
        }

        return new String(canonical, 0, j);
    }
}
