/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.util.LinkedList;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;
import org.globus.wsrf.ResourceKey;
import org.globus.exec.utils.Resources;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.factory.FactoryHomeConfiguration;

class RunQueue
{
    /** Log4J logger */
    private static Log logger = LogFactory.getLog(RunQueue.class);
    private static I18n i18n = I18n.getI18n(Resources.class.getName());
    private static RunQueue RUN_QUEUE;
    private static final int MINIMUM_RUN_THREAD_COUNT = 1;
    private static final int DEFAULT_RUN_THREAD_COUNT = 10;
    private LinkedList readyQueue;

    static
    {
        // get the configured number of run queues
        FactoryHomeConfiguration homeConfig
            = ManagedJobFactoryHome.getConfiguration();
        int runThreadCount;

        // read the number of run threads from the container registry
        try
        {
            runThreadCount
                = Integer.parseInt(homeConfig.getRunQueueThreadCount());
        }
        catch (Exception e)
        {
            logger.warn(i18n.getMessage(Resources.BAD_RUN_QUEUE_THREAD_COUNT),
                        e);
            runThreadCount = DEFAULT_RUN_THREAD_COUNT;
        }

        // assert that there isn't a silly setting
        if (runThreadCount < MINIMUM_RUN_THREAD_COUNT)
        {
            logger.warn(i18n.getMessage(Resources.BAD_RUN_QUEUE_THREAD_COUNT));
            runThreadCount = DEFAULT_RUN_THREAD_COUNT;
        }

        // create the single RunQueue object
        RUN_QUEUE = new RunQueue();          

        // create run thread instances
        for (int threadIndex=0; threadIndex<runThreadCount; threadIndex++)
        {
            RunThread runThread = new RunThread(
                "RunQueueThread_" + threadIndex, RUN_QUEUE);
            runThread.setDaemon(true);
            runThread.start();
        }
        logger.debug("created RunQueue with " + runThreadCount + " threads");
    }

    // private constructor: called during initialization in static block
    private RunQueue()
    { 
        this.readyQueue = new LinkedList();
    }

    public synchronized void add(ResourceKey resourceKey)
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("Entering add()");
        }

        this.readyQueue.add(resourceKey);
        notify();

        if (logger.isDebugEnabled())
        {
            logger.debug("adding " + resourceKey.toString() + " to queue.");
            printRunQueueGraph();
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving add()");
        }
    }

    public synchronized ResourceKey remove()
    {
        ResourceKey resourceKey = (ResourceKey) this.readyQueue.removeFirst();

        if (logger.isDebugEnabled())
        {
            logger.debug("removing " + resourceKey.toString() + " from queue.");
            printRunQueueGraph();
        }

        return resourceKey;
    }

    private synchronized void printRunQueueGraph() 
    {
        int queuedJobCount = this.readyQueue.size();
        StringBuffer asterisks = new StringBuffer();
        for (int index=0; index<queuedJobCount; index++)
        {
            asterisks.append("*");
        }
        logger.debug("\nRunQueue Graph: " + asterisks.toString());
    }

    public synchronized int size()
    {
        return this.readyQueue.size();
    }

    public static RunQueue getInstance()
    {
        return RUN_QUEUE;
    }
}
