/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.StringTokenizer;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.PositiveInteger;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.InternalStateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.JobIdStatePairType;
import org.globus.exec.generated.JobTypeEnumeration;
import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.ScriptCommandEnumeration;
import org.globus.exec.generated.StagingTypeEnumeration;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.service.ManagedJobHelper;
import org.globus.exec.service.exec.utils.FileMapping;
import org.globus.exec.service.exec.utils.StagingHelper;
import org.globus.exec.service.exec.utils.LocalStagingHelper;
import org.globus.exec.service.exec.utils.WSStagingHelper;
import org.globus.exec.service.factory.FactoryServiceConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.factory.ManagedJobFactoryResource;
import org.globus.exec.service.factory.ManagedJobFactoryService;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.usage.GramUsageMonitorPacket;
import org.globus.exec.service.utils.AuditDatabaseConfiguration;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.generated.ServiceLevelAgreementType;
import org.globus.exec.monitoring.AlreadyRegisteredException;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.audit.AuditUtil;
import org.globus.exec.utils.audit.AuditDatabaseAppender;
import org.globus.gram.internal.GRAMProtocolErrorConstants;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.RepeatedlyStartedFaultType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.util.GlobusURL;
import org.globus.util.I18n;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.container.UsageConfig;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.PerformanceLog;
import org.globus.wsrf.utils.XmlUtils;
import org.oasis.wsrf.faults.BaseFaultType;
import org.w3c.dom.Element;


//TODO add JobManager fault creation methods to FaultUtils

//TODO finish populating the state transition methods

/*TODO in transition methods which call scripts, wait() on the JobManagerScript
       instance. Have JobManagerScript call notify() on itself when it's done.
*/

/*TODO create extension of MEJR that overrides [s/g]etResourceProperty and
       [s/g]etResourceDatum methods for testing by setting fake values.
*/

/*TODO create extensions of StateMachine which override transition methods
       for simulating behavior in unit tests.
*/

public class StateMachine {
    static I18n i18n = I18n.getI18n(Resources.class.getName());

    /** Log4J logger */
    static Log logger = LogFactory.getLog(StateMachine.class);

    /** Performance Logger */
    private static PerformanceLog performanceLogger = new PerformanceLog(
        StateMachine.class.getName() + ".performance");

    /** Audit Logger */
    private static Log auditLogger = LogFactory.getLog(
        StateMachine.class.getName() + ".audit");

    /** Usage Config */
    private static UsageConfig usageConfig;

    /** Staging helper that will be used for staging */
    private static StagingHelper stagingHelper = null;

    private StateMachine() { }
    
    static void init()
    {
        usageConfig = UsageConfig.getUsageConfig();
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        
        boolean localInvocations
            = (new Boolean (
                 factoryHomeConfiguration
                   .getEnableLocalInvocations()))
                     .booleanValue();

        logger.debug("local invocations: "+localInvocations);

        if (localInvocations) {
            stagingHelper = new LocalStagingHelper();
        } else {
            stagingHelper = new WSStagingHelper();
        }
    }

    static void processState(ResourceKey resourceKey)
        throws                              ResourceException
    {
        if (logger.isTraceEnabled())
        {
            logger.debug(getLoggingPrefix(resourceKey) +
                "Entering processState()");
        }

        //resolve resource key
        if (logger.isTraceEnabled())
        {
            logger.debug(getLoggingPrefix(resourceKey) +
                "Finding resource reference for key "
                + resourceKey.toString());
        }
        ManagedExecutableJobResource resource = null;
        ManagedExecutableJobHome home = null;
        try {
            home = ManagedExecutableJobHome.getHome();
            resource
                = (ManagedExecutableJobResource) home.unsyncFind(resourceKey);
        } catch (ResourceException re)
        {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_FIND_ERROR,
                new String[] { resourceKey.getValue().toString() });

            logger.debug(errorMessage);
            return;
        }
        if (logger.isTraceEnabled())
        {
            logger.debug(getLoggingPrefix(resourceKey) +
                "Found resource reference for key "
                + resourceKey.toString());
        }

        synchronized (resource.getStateTransitionMutex()) {
            //Get current internal state
            InternalStateEnumeration internalState
                = resource.getInternalState();

            //ignore state transition request if Done or Failed
            StateEnumeration state = resource.getState();
            if (   !internalState.equals(InternalStateEnumeration.Restart)
                && (   state.equals(StateEnumeration.Done)
                    || state.equals(StateEnumeration.Failed)))
            {
                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resourceKey) + 
                        "Ignoring processState() since job is done");
                }
                return;
            }

            //Check for UserCancel or SystemCancel request
            boolean canceled = ((Boolean) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CANCELED))
                .booleanValue();
            boolean userCancelRequested = ((Boolean) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_USER_CANCEL_REQUESTED))
                .booleanValue();
            boolean systemCancelRequested
                = ((Boolean) resource.getResourceDatum(
                    ManagedExecutableJobConstants.RD_SYSTEM_CANCEL_REQUESTED))
                .booleanValue();
            if (!canceled && (userCancelRequested || systemCancelRequested))
            {
                //select correct transition method for cancel request
                if (   internalState.equals(
                        InternalStateEnumeration.Done)
                    || internalState.equals(
                        InternalStateEnumeration.Failed))
                {
                    logger.debug(getLoggingPrefix(resourceKey) + 
                        "Ignoring cancel request for terminal state "
                        + internalState);
                }
                else
                if (   internalState.equals(
                        InternalStateEnumeration.FileCleanUp)
                    || internalState.equals(
                        InternalStateEnumeration.FileCleanUpResponse)
                    || internalState.equals(
                        InternalStateEnumeration.FailureFileCleanUp)
                    || internalState.equals(
                        InternalStateEnumeration.FailureFileCleanUpResponse))
                {
                    logger.debug(getLoggingPrefix(resourceKey) + 
                        "Setting internal state to FailureCacheCleanUp");
                    resource.setInternalState(
                        InternalStateEnumeration.FailureCacheCleanUp);
                    internalState
                        = InternalStateEnumeration.FailureCacheCleanUp;
                }
                else
                if (   internalState.equals(
                        InternalStateEnumeration.None)
                    || internalState.equals(
                        InternalStateEnumeration.CacheCleanUp)
                    || internalState.equals(
                        InternalStateEnumeration.FailureCacheCleanUp))
                {
                    logger.debug(getLoggingPrefix(resourceKey) + "Setting internal state to Failed");
                    resource.setInternalState(InternalStateEnumeration.Failed);
                    internalState
                        = InternalStateEnumeration.Failed;
                }
                else
                if (userCancelRequested)
                {
                    logger.debug(getLoggingPrefix(resourceKey) + "Setting internal state to UserCancel");
                    resource.setInternalState(
                        InternalStateEnumeration.UserCancel);
                    internalState = InternalStateEnumeration.UserCancel;
                }
                else {
                    logger.debug(getLoggingPrefix(resourceKey) + 
                        "Setting internal state to SystemCancel from "
                        + internalState);
                    resource.setInternalState(
                        InternalStateEnumeration.SystemCancel);
                    internalState = InternalStateEnumeration.SystemCancel;
                }

                //set canceled flag so we don't do this more than once
                resource.setResourceDatum(
                    ManagedExecutableJobConstants.RD_CANCELED,
                    true);
            }

            if (logger.isDebugEnabled())
            {
                logger.debug(getLoggingPrefix(resourceKey) +
                    "processing internal state " + internalState);
                StringBuffer logMsg = new StringBuffer(
                    "\n------------------------------------------------------")
                .append("\nPROCESSING INTERNAL STATE:  ")
                .append("-- " + internalState + " --")
                .append(
                    "\n------------------------------------------------------");
                logger.debug(getLoggingPrefix(resourceKey) + logMsg.toString());
            }

            //Check for procStateChanged flag
            if (isNotifiedJobData(resource))
            {
                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resourceKey)
                                + "Handling job process state change");
                }

                //ignore state change message if we're not expecting it
                if (!internalState.equals(
                        InternalStateEnumeration.WaitingForStateChanges))
                {
                    if (   state.equals(StateEnumeration.StageOut)
                        || state.equals(StateEnumeration.CleanUp)
                        || state.equals(StateEnumeration.Done)
                        || state.equals(StateEnumeration.Failed))
                    {
                        if (logger.isDebugEnabled()) {
                            logger.debug(getLoggingPrefix(resourceKey) + 
                                "Not waiting for state changes anymore"
                                +"--ignoring.");
                        }

                        removeNotifiedJobData(resource);
                    }
                }
            }

            //Determine the transition method of the current internal state
            String transitionMethodName
                = "process" + internalState.toString() + "State";
            Method internalStateTransitionMethod = null;
            try {
                internalStateTransitionMethod
                    = StateMachine.class.getDeclaredMethod(
                        transitionMethodName,
                        new Class[] { ManagedExecutableJobResource.class });
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.STATE_TRANSITION_METHOD_NOT_FOUND,
                    transitionMethodName);
                throw new RuntimeException(errorMessage, e);
            }

            //Invoke the transition method
            try {
                internalStateTransitionMethod.invoke(
                    null, new Object[] { resource });
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.STATE_TRANSITION_METHOD_ERROR,
                    transitionMethodName);
                throw new RuntimeException(errorMessage, e);
            }
        }

        if (logger.isTraceEnabled())
        {
            logger.debug(
                getLoggingPrefix(resourceKey) + "Leaving processState()");
        }
    }


    //---------- State Transition Interrupt Methods ----------//
    static void userCancel(
        ManagedExecutableJobResource        resource)
    {
        if (resource.getFault() == null)
        {
            FaultType fault = FaultUtils.makeFault(
                FaultType.class,
                (ResourceKey) resource.getID(),
                i18n.getMessage(Resources.USER_CANCEL),
                null,
                resource.getState(),
                null,
                0);
            resource.setFault(fault);
        }
        else
        {
            Exception cause = new Exception(
                i18n.getMessage(Resources.USER_CANCEL));
            cause.fillInStackTrace();
            logger.error(i18n.getMessage(Resources.SECONDARY_FAULT), cause);
        }

        boolean started = ((Boolean) resource.getResourceDatum(
            ManagedJobConstants.RD_STARTED))
            .booleanValue();

        cancel(
            resource,
            ManagedExecutableJobConstants.RD_USER_CANCEL_REQUESTED,
            started);
    }

    public static void systemCancel(
        ManagedExecutableJobResource        resource,
        String                              errorMessage,
        Exception                           cause)
    {
        FaultType fault = FaultUtils.makeFault(
            FaultType.class,
            (ResourceKey) resource.getID(),
            errorMessage,
            cause,
            resource.getState(),
            null,
            0);

        resource.setFault(fault);

        systemCancel(resource);
    }

    static void systemCancel(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled())
        {
            logger.debug("System cancel requested for job "
                        +((ResourceKey)resource.getID()).getValue());
            try {
                throw new Exception();
            } catch (Exception e) {
                logger.debug("systemCancel stack trace", e);
            }
        }

        boolean started = ((Boolean) resource.getResourceDatum(
            ManagedJobConstants.RD_STARTED))
            .booleanValue();

        cancel(
            resource,
            ManagedExecutableJobConstants.RD_SYSTEM_CANCEL_REQUESTED,
            started);
    }

    private static void cancel(
        ManagedExecutableJobResource        resource,
        String                              cancelFlagName,
        boolean                             addToRunQueue)
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("Entering cancel()");
        }

        synchronized (resource.getStateTransitionMutex())
        {
            boolean canceled = ((Boolean) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CANCELED))
                .booleanValue();
            if (!canceled)
            {
                resource.setResourceDatum(cancelFlagName, true);

                if (addToRunQueue)
                {
                    if (logger.isDebugEnabled()) {
                        logger.trace("Adding to RunQueue.");
                    }
                    RunQueue.getInstance().add((
                        ResourceKey)resource.getID());
                }
            }
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving cancel()");
        }
    }

    static void localJobStateChanged(
        ManagedExecutableJobResource        resource,
        String                              jobId,
        Date                                timeStamp,
        StateEnumeration                    state,
        int                                 exitCode)
    {
        synchronized (resource.getStateTransitionMutex())
        {
            insertNotifiedJobData(resource, jobId, state);

            resource.setResourceProperty(
                ManagedExecutableJobConstants.RP_EXIT_CODE,
                new Integer(exitCode));


            if (logger.isDebugEnabled()) {
                logger.trace("Adding to RunQueue.");
            }
            RunQueue.getInstance().add((
                ResourceKey)resource.getID());
        }
    }

    static void release(
        ManagedExecutableJobResource        resource)
    {
        synchronized (resource.getStateTransitionMutex())
        {
            boolean holding = ((Boolean) resource.getResourceProperty(
                ManagedJobConstants.RP_HOLDING)).booleanValue();

            //release any existing hold
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                false);

            //pre-release any future hold by preventing a match to any state
            resource.setResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE,
                null);

            if (holding)
            {
                if (logger.isDebugEnabled()) {
                    logger.trace("Adding to RunQueue.");
                }
                RunQueue.getInstance().add((
                    ResourceKey)resource.getID());
            }
        }
    }
    //------------------------------------------------//


    //--------------- Transition Methods ---------------//
    private static void processUnsubmittedHoldState(
        ManagedExecutableJobResource        resource)
    {
        StateEnumeration state = resource.getState();
        resource.setFault(
            FaultUtils.createUnsupportedFeatureFault(
                "Unsubmitted", "holdState", state,
                null, -1, (ResourceKey) resource.getID()));
        resource.setResourceProperty(
            ManagedJobConstants.RP_HOLDING,
            false);

        systemCancel(resource);
    }

    private static void processNoneState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processNoneState

        resource.setResourceDatum(
            ManagedJobConstants.RD_STARTED,
            new Boolean(true));

        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null)
            && holdState.equals(StateEnumeration.Unsubmitted)) {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.UnsubmittedHold, true);

            performanceLogger.stop("processNoneState");

            return;
        }

        processNewState(resource, InternalStateEnumeration.Start, true);

        performanceLogger.stop("processNoneState");
    }

    private static void processStartState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processStartState

        try
        {
            logJobAccepted(resource);
        }
        catch (Exception e)
        {
            //TODO i18n this
            logger.error("Failed to log accepted job message.");
        }

        JobDescriptionType jobDescription = resource.getJobDescription();
        Object fileStageIn = jobDescription.getFileStageIn();
        Object holdState = jobDescription.getHoldState();
        if (   (fileStageIn != null)
            || (   (holdState != null)
                && (holdState.equals(StateEnumeration.StageIn))))
        {
            logger.debug(getLoggingPrefix(resource) +
                "Not skipping StageIn");
            processNewState(resource, InternalStateEnumeration.StageIn, true);
        }
        else
        {
            logger.debug(getLoggingPrefix(resource) +
                "Skipping StageIn and going straight to Submit");
            processNewState(resource, InternalStateEnumeration.Submit, true);
        }

        logAuditRecord(true, resource, false, false);
        performanceLogger.stop("processStartState");
    }

    private static void processRestartState(
        ManagedExecutableJobResource        resource)
    {
        logger.debug("Entering processRestartState()");

        //get the internal state the job was in before the restart
        InternalStateEnumeration restartInternalState
            = (InternalStateEnumeration) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_RESTART_INTERNAL_STATE);

        //re-register transfer job to the staging listener
        EndpointReferenceType  cleanUpEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CLEAN_UP_ENDPOINT);
        EndpointReferenceType stageOutEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_STAGE_OUT_ENDPOINT);
        EndpointReferenceType stageInEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_STAGE_IN_ENDPOINT);
        EndpointReferenceType transferJobEndpoint = null;
        
        if (   cleanUpEndpoint != null
            && (restartInternalState.equals(InternalStateEnumeration.FileCleanUp)
                || restartInternalState.equals(InternalStateEnumeration.FileCleanUpResponse)))
        {
            transferJobEndpoint = cleanUpEndpoint;
        }
        else
        if (   stageOutEndpoint != null 
            && (restartInternalState.equals(InternalStateEnumeration.StageOut)
                || restartInternalState.equals(InternalStateEnumeration.StageOutResponse)))
        {
            transferJobEndpoint = stageOutEndpoint;
        }
        else
        if (   stageInEndpoint != null
            && (restartInternalState.equals(InternalStateEnumeration.StageIn)
                || restartInternalState.equals(InternalStateEnumeration.StageInResponse)))
        {
            transferJobEndpoint = stageInEndpoint;
        }

        if (transferJobEndpoint != null)
        {
            /*Attempt to start the transfer job. Either it will fail with a
             *RepeatedlyStartedFaultType (after re-registering with the staging
             *listener) which we can ignore, or it will start the transfer job
             *as we intended to do before the container crash.
             */
            try {
                logger.debug("attempting to start recovered transfer job");
                stagingHelper.startTransferJob(resource, transferJobEndpoint);
            }
            catch (RepeatedlyStartedFaultType rsft)
            {
                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resource)
                        + "ignoring RepeatedlyStartedFaultType from RFT");
                }
            }
            catch (Exception e)
            {
                StagingTypeEnumeration stagingType = null;
                if (   restartInternalState.equals(
                        InternalStateEnumeration.StageIn)
                    || restartInternalState.equals(
                        InternalStateEnumeration.StageInResponse))
                {
                    stagingType = StagingTypeEnumeration.StageIn;
                }
                else
                if (   restartInternalState.equals(
                        InternalStateEnumeration.StageOut)
                    || restartInternalState.equals(
                        InternalStateEnumeration.StageOutResponse))
                {
                    stagingType = StagingTypeEnumeration.StageOut;
                }
                else
                {
                    stagingType = StagingTypeEnumeration.CleanUp;
                }

                resource.setFault(
                    createStagingFault(
                        resource,
                        stagingType,
                        e));

                logger.trace(
                    getLoggingPrefix(resource) + "calling systemCancel()");
                systemCancel(resource);
            }
        }

        //re-subscribe for state notifications
        String job_ids[] = (String []) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
        if (job_ids != null) {
            try {
               subscribeAtJobStateMonitor(resource);
            } catch (AlreadyRegisteredException e) {
                  String errorMessage = i18n.getMessage(
                              Resources.JOB_ID_REGISTRATION_ERROR);
                  logger.warn("Register attempt failed for job "+
                              ((ResourceKey) resource.getID()).getValue()); 
                  systemCancel(resource, errorMessage, e);
                  return;
            }
        }

        //reconstitute the job credential listener
        String localUserId
            = (String) resource.getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID);
        File userGlobusDirectory
            = ManagedExecutableJobResource.getUserGlobusDirectory(localUserId);
        JobDescriptionType jobDescription = resource.getJobDescription();
        if (jobDescription.getJobCredentialEndpoint() != null)
        {
            try
            {
                resource.getJobCredential(userGlobusDirectory);
            } catch (Exception e)
            {
                StateEnumeration state = resource.getState();
                if (   !state.equals(StateEnumeration.Done)
                    && !state.equals(StateEnumeration.Failed))
                {
                    String errorMessage = i18n.getMessage(
                        Resources.CREDENTIAL_ACQUISITION_ERROR);
                    FaultType fault = FaultUtils.makeFault(
                        FaultType.class,
                        (ResourceKey) resource.getID(),
                        errorMessage,
                        e,
                        resource.getState(),
                        null,
                        0);
                    resource.setFault(fault);

                    systemCancel(resource);
                }
            }
        }

        //reconstitute the staging credential listener
        if (jobDescription.getStagingCredentialEndpoint() != null)
        {
            try
            {
                resource.getStagingCredential();
            } catch (Exception e)
            {
                StateEnumeration state = resource.getState();
                if (   !state.equals(StateEnumeration.Done)
                    && !state.equals(StateEnumeration.Failed))
                {
                    String errorMessage = i18n.getMessage(
                        Resources.CREDENTIAL_ACQUISITION_ERROR);
                    FaultType fault = FaultUtils.makeFault(
                        FaultType.class,
                        (ResourceKey) resource.getID(),
                        errorMessage,
                        e,
                        resource.getState(),
                        null,
                        0);
                    resource.setFault(fault);

                    systemCancel(resource);
                }
            }
        }

        logger.debug("processRestartState with restartInternalState == " 
                + restartInternalState.toString());


        processNewState(resource, restartInternalState, true);

        logger.debug("Leaving processRestartState()");
    }

    private static void processStageInHoldState(
        ManagedExecutableJobResource        resource)
    {
        boolean holding = ((Boolean) resource.getResourceProperty(
            ManagedJobConstants.RP_HOLDING)).booleanValue();
        if (!holding)
        {
            processNewState(resource, InternalStateEnumeration.StageIn, true);
        } else
        {
            JobDescriptionType jobDescription
                = ((ServiceLevelAgreementType) resource.getResourceProperty(
                    ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
            if (jobDescription.getFileStageIn() == null) {
                resource.setFault(
                    createStagingHoldFault(
                        resource,
                        StagingTypeEnumeration.StageIn));

                resource.setResourceProperty(
                    ManagedJobConstants.RP_HOLDING,
                    false);

                systemCancel(resource);
                return;
            }

            try {
                resource.setState(StateEnumeration.StageIn);
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    private static void processStageInState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processStageInSTate

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource)
                + "Entering processStageInState()");
        }

        //catch stageInHold
        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null)
            && holdState.equals(StateEnumeration.StageIn))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);
            processNewState(
                resource, InternalStateEnumeration.StageInHold, true);

            return;
        }

        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        TransferRequestType fileStageIn = jobDescription.getFileStageIn();

        boolean staging = false;
        if (fileStageIn != null)
        {
            if (logger.isDebugEnabled()) {
                logger.debug(getLoggingPrefix(resource) +
                    "Processing fileStageIn directives.");
            }

            staging = true;

            logger.trace(getLoggingPrefix(resource)
                        + "calling MEJR.setState()");
            try {
                resource.setState(StateEnumeration.StageIn);
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }

            //submit a transfer request to RFT
            try {
                logger.trace(getLoggingPrefix(resource) +
                    "calling StagingHelper.submitStagingRequest()");
                EndpointReferenceType stageInEndpoint
                    = stagingHelper.submitStagingRequest(resource, fileStageIn);
                resource.setResourceDatum(
                    ManagedExecutableJobConstants.RD_STAGE_IN_ENDPOINT,
                    stageInEndpoint);

                logger.trace(getLoggingPrefix(resource) +
                    "calling MEJR.setResourceDatum()");
            } catch (Exception e) {
                resource.setFault(
                    createStagingFault(
                        resource,
                        StagingTypeEnumeration.StageIn,
                        e));

                logger.trace(
                    getLoggingPrefix(resource) + "calling systemCancel()");
                systemCancel(resource);

                performanceLogger.stop("processStageInState");

                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resource) +
                        "Leaving processStageInState()");
                }
                return;
            }
        }

        if (staging)
        {
            logger.trace(getLoggingPrefix(resource) +
                "calling processNewState(StageInResponse)");
            processNewState(
                resource, InternalStateEnumeration.StageInResponse, false);
        } else
        {
            logger.trace(getLoggingPrefix(resource) +
                "calling processNewState(Submit)");
            processNewState(resource, InternalStateEnumeration.Submit, true);
        }

        if (logger.isDebugEnabled()) {
            logger.debug(
                getLoggingPrefix(resource) + "Leaving processStageInState()");
        }

        performanceLogger.stop("processStageInState");
    }

    private static void processStageInResponseState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processStageInResponseState()");
        }

        try {
            String source = null;
            String destination = null;
            String attribute = "fileStageIn";
            StateEnumeration state
                = (StateEnumeration) resource.getResourceProperty(
                    ManagedJobConstants.RP_STATE);

            EndpointReferenceType transferJobEndpoint
                = (EndpointReferenceType) resource.getResourceDatum(
                    ManagedExecutableJobConstants.RD_STAGE_IN_ENDPOINT);
            if (transferJobEndpoint == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.warn("Missing transfer job endpoint. Continuing...");
                }

                processNewState(
                    resource, InternalStateEnumeration.Submit, true);

                return;
            }

            RequestStatusType status
                = StagingListener.getNextStatus(transferJobEndpoint);
            if (status == null)
            {
                status = stagingHelper.poll(resource,transferJobEndpoint);
                if (status == null)
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Missing RFT status. "
                                    + "Probably an out-of-order notification.");
                    }
                    return;
                }
            }

            //handle failed stage in
            if (status.getRequestStatus().equals(
                    RequestStatusTypeEnumeration.Failed))
            {
                if (logger.isDebugEnabled()) {
                    logger.debug("Detected a StageIn transfer failure");
                }

                BaseFaultType transferFault
                    = stagingHelper.getTransferFaultFromRP(status.getFault());

                //set fault RP if not already set
                if (resource.getFault() == null)
                {
                    if (logger.isDebugEnabled()) {
                        logger.debug("A fault does not exist, so create one.");
                    }
                    FaultType fault = null;
                    if (transferFault != null) {
                        fault = FaultUtils.createStagingFault(
                            source,
                            destination,
                            attribute,
                            resource.getState(),
                            null,
                            (ResourceKey) resource.getID(),
                            transferFault);
                    } else {
                        fault = FaultUtils.createStagingFault(
                            source,
                            destination,
                            attribute,
                            resource.getState(),
                            null,
                            (ResourceKey) resource.getID());
                    }

                    resource.setFault(fault);
                }


                try {
                    stagingHelper.destroyTransferJob(
                        resource, transferJobEndpoint);
                } catch(Exception e) {
                    String errorMessage = i18n.getMessage(
                        Resources.RFT_RESOURCE_DESTROY_ERROR);
                    logger.error(errorMessage, e);
                }

                StateMachine.systemCancel(resource);

                return;
            }

            if (!processCompletedStagingJob(
                resource, status, InternalStateEnumeration.Submit,
                transferJobEndpoint))
            {
                systemCancel(resource);
            }
        }
        catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                            Resources.STATE_NOTIFICATION_PROCESSING_ERROR);
            logger.error(errorMessage, e);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving processStageInResponseState()");
        }
    }

    private static void processPendingHoldState(
        ManagedExecutableJobResource        resource)
    {
        try {
            resource.setState(StateEnumeration.Pending);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }

        boolean holding = ((Boolean) resource.getResourceProperty(
            ManagedJobConstants.RP_HOLDING)).booleanValue();
        if (!holding) {
            processNewState(resource, InternalStateEnumeration.Submit, true);
        }
    }

    private static void processActiveHoldState(
        ManagedExecutableJobResource        resource)
    {
        resource.setFault(
            FaultUtils.createUnsupportedFeatureFault(
                "Active", "holdState", resource.getState(),
                null, -1, (ResourceKey) resource.getID()));
        resource.setResourceProperty(
            ManagedJobConstants.RP_HOLDING,
            false);

        //processNewState(resource, InternalStateEnumeration.SystemCancel, true);
        systemCancel(resource);
    }

    private static void processSubmitState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processSubmitState

        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.Pending))
        {
            if (logger.isDebugEnabled()) {
                logger.trace("Initiating hold on " + holdState);
            }

            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.PendingHold, true);

            performanceLogger.stop("processSubmitState");

            return;
        }

        String localResourceManager = (String) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER);
        PerlJobDescription perlJobDescription
            = resource.getPerlJobDescription();

        if (localResourceManager.equals(
            ManagedJobFactoryConstants.FACTORY_TYPE.FORK)) {
            perlJobDescription.put("useforkstarter", "1");
        }
        perlJobDescription.put("expand_globus_home", "1");
        perlJobDescription.put("expand_globus_location", "1");

        JobManagerScript script = null;
        try
        {
            script = runScript(resource, ScriptCommandEnumeration.submit);
        }
        catch (Exception e)
        {
            if (script != null)
            {
                resource.setFault(
                    createFaultFromErrorCode(
                        resource,
                        script.getCommand(),
                        0,
                        script.getFailureMessage()));
                StateMachine.systemCancel(resource);
            }
            else
            {
                String errorMessage = i18n.getMessage(
                    Resources.SCRIPT_EXECUTION_ERROR);
                FaultType fault = FaultUtils.makeFault(
                    FaultType.class,
                    (ResourceKey) resource.getID(),
                    errorMessage,
                    e,
                    resource.getState(),
                    null,
                    0);
                resource.setFault(fault);
                StateMachine.systemCancel(resource);
                return;
            }
        }

        //wait for script to finish
        synchronized (script)
        {
            while (!script.isDone())
            {
                try {
                    script.wait(2000);
                } catch (Exception e) {
                    String errorMessage =
                        i18n.getMessage(Resources.INTERRUPTED_JOB_SCRIPT_WAIT,
                                        script.getCommand());
                    logger.error(errorMessage, e);
                }
            }
        }
        if (logger.isDebugEnabled()) {
            logger.trace("Done waiting for " + script.getCommand()
                        + " script");
        }

        //process script output
        StateEnumeration newState = script.getJobState();
        int errorCode = script.getError();
        if (logger.isDebugEnabled()) {
            logger.debug("script return code: " + errorCode);
        }

        if (errorCode != 0)
        {
            logger.trace("script return code means error!");
            resource.setFault(
                createFaultFromErrorCode(
                    resource,
                    script.getCommand(),
                    errorCode,
                    script.getFailureMessage()));

            systemCancel(resource);
        }
        else
        if ((script.getJobId() == null) || script.getJobId().equals(""))
        {
            if (logger.isDebugEnabled()) {
                logger.debug("script returned no job ID!");
                logger.trace("Job Submission Failed");
            }
            resource.setFault(
                createFaultFromErrorCode(
                    resource,
                    script.getCommand(),
                    GRAMProtocolErrorConstants.INVALID_SCRIPT_REPLY,
                    script.getFailureMessage()));

            systemCancel(resource);
        }
        else
        {
            //save the time at which we know the local job was queued
            resource.setResourceDatum(
                ManagedExecutableJobConstants.RD_QUEUED_TIME,
                new Date());

            //catch Active hold
            if (  (holdState != null)
                && holdState.equals(StateEnumeration.Active))
            {
                resource.setResourceProperty(
                    ManagedJobConstants.RP_HOLDING,
                    true);
                processNewState(
                    resource,
                    InternalStateEnumeration.ActiveHold,
                    true);

                performanceLogger.stop("processSubmitState");

                return;
            }

            if (logger.isDebugEnabled()) {
                logger.debug("script returned job state: " +
                             newState.getValue());
            }

            if (newState.equals(StateEnumeration.Done))
            {
                processNewState(
                    resource, InternalStateEnumeration.MergeStdout, true);
            } else
            if (newState.equals(StateEnumeration.Failed))
            {
                systemCancel(resource);
            } else
            {
                //save job ID
                String localJobId = script.getJobId();
                perlJobDescription.put("jobid", localJobId);
                setLocalJobId(resource, localJobId, newState);

                //log that the job was submitted
                if (localJobId != null)
                {
                    try
                    {
                        logJobSubmitted(resource, localJobId);
                    }
                    catch (Exception e)
                    {
                        //TODO i18n this
                        logger.error("Failed to log submitted job message.", e);
                    }
                }

                //subscribe for state notifications
                logger.debug("Subscribing job " + resource.getID() +
                             " to SEG events for local " +
                             "job ID " + localJobId);
                try {
                   subscribeAtJobStateMonitor(resource);
                } catch (AlreadyRegisteredException e) {
                    String errorMessage = i18n.getMessage(
                                Resources.JOB_ID_REGISTRATION_ERROR);
                    logger.warn("Register attempt failed for job "+
                                ((ResourceKey) resource.getID())
                                .getValue().toString()); 
                    systemCancel(resource, errorMessage, e);
                    return;
                }

                processNewState(
                    resource,
                    InternalStateEnumeration.WaitingForStateChanges,
                    false);
            }
        }

        logAuditRecord(false, resource, false, false);
        performanceLogger.stop("processSubmitState");
    }

    private static void processOpenStdoutState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processOpenStdoutState

        //determine stdout location
        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        String stdout = jobDescription.getStdout();

        //setup the stdoutUrl resource property if appropriate
        if ((stdout != null) && (!stdout.equals("/dev/null")))
        {
            String stdoutUrl = null;
            FileMapping fileMapping = null;
            GlobusURL localStdoutUrl = null;
            Exception cause = null;
            try
            {
                fileMapping = resource.getFileSystemMapping();

                localStdoutUrl = new GlobusURL(
                    "file:///" + stdout);

                stdoutUrl = fileMapping.mapFileName(localStdoutUrl);
            }
            catch(MalformedURLException murle)
            {
                String errorMessage = i18n.getMessage(
                    Resources.INVALID_OUTPUT_URL,
                    new String[] {"stdout", stdout});
                cause = new Exception(errorMessage, murle);
                cause.fillInStackTrace();
            }
            catch(Exception e)
            {
                cause = e;
            }

            if (cause != null)
            {
                if (resource.getFault() != null)
                {
                    FaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "stdout",
                            resource.getState(),
                            null,
                            0,
                            (ResourceKey) resource.getID(),
                            cause);
                    resource.setFault(fault);
                }
                else
                {
                    logger.error(
                        i18n.getMessage(Resources.SECONDARY_FAULT),
                        cause);
                }

                systemCancel(resource);
                return;
            }

            logger.trace("stdout URL " + stdoutUrl);

            //set stdout GridFTP URL
            try
            {
                resource.setResourceProperty(
                    ManagedExecutableJobConstants.RP_STDOUT_URL,
                    new URI(stdoutUrl));
            }
            catch (Exception e)
            {
                if (resource.getFault() != null)
                {
                    String errorMessage = i18n.getMessage(
                        Resources.RP_SET_ERROR,
                        new String[] {
                            ManagedExecutableJobConstants.RP_STDOUT_URL
                            .getLocalPart(),
                            stdoutUrl});
                    cause = new Exception(errorMessage);
                    cause.fillInStackTrace();
                    FaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "stdout",
                            resource.getState(),
                            null,
                            0,
                            (ResourceKey) resource.getID(),
                            cause);
                    resource.setFault(fault);
                }
                else
                {
                    logger.error(
                        i18n.getMessage(Resources.SECONDARY_FAULT),
                        e);
                }

                systemCancel(resource);
            }
        }

        processNewState(
            resource, InternalStateEnumeration.OpenStderr, true);

        performanceLogger.stop("processOpenStdoutState");
    }

    private static void processOpenStderrState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processOpenStderrState

        //determine stderr location
        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        String stderr = jobDescription.getStderr();

        //setup the stdoutUrl resource property if appropriate
        if ((stderr != null) && (!stderr.equals("/dev/null")))
        {
            String stderrUrl = null;
            FileMapping fileMapping = null;
            GlobusURL localStderrUrl = null;
            Exception cause = null;
            try
            {
                fileMapping = resource.getFileSystemMapping();

                localStderrUrl = new GlobusURL(
                    "file:///" + stderr);

                stderrUrl = fileMapping.mapFileName(localStderrUrl);
            }
            catch(MalformedURLException murle)
            {
                String errorMessage = i18n.getMessage(
                    Resources.INVALID_OUTPUT_URL,
                    new String[] {"stderr", stderr});
                cause = new Exception(errorMessage, murle);
                cause.fillInStackTrace();
            }
            catch(Exception e)
            {
                cause = e;
            }

            if (cause != null)
            {
                if (resource.getFault() != null)
                {
                    FaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "stderr",
                            resource.getState(),
                            null,
                            0,
                            (ResourceKey) resource.getID(),
                            cause);
                    resource.setFault(fault);

                    systemCancel(resource);
                }
                else
                {
                    logger.error(
                        i18n.getMessage(Resources.SECONDARY_FAULT),
                        cause);
                }

                return;
            }

            logger.trace("stderr URL " + stderrUrl);

            //set stderr GridFTP URL
            try
            {
                resource.setResourceProperty(
                    ManagedExecutableJobConstants.RP_STDERR_URL,
                    new URI(stderrUrl));
            }
            catch (Exception e)
            {
                if (resource.getFault() != null)
                {
                    String errorMessage = i18n.getMessage(
                        Resources.RP_SET_ERROR,
                        new String[] {
                            ManagedExecutableJobConstants.RP_STDERR_URL
                            .getLocalPart(),
                            stderrUrl});
                    cause = new Exception(errorMessage);
                    cause.fillInStackTrace();
                    FaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "stderr",
                            resource.getState(),
                            null,
                            0,
                            (ResourceKey) resource.getID(),
                            cause);
                    resource.setFault(fault);

                    systemCancel(resource);
                }
                else
                {
                    logger.error(
                        i18n.getMessage(Resources.SECONDARY_FAULT),
                        e);
                }

                return;
            }
        }

        try {
            resource.setState(StateEnumeration.Active);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }


        processNewState(
            resource, InternalStateEnumeration.WaitingForStateChanges, false);

        performanceLogger.stop("processOpenStderrState");
    }

    private static void processWaitingForStateChangesState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processWaitingForStateChangesState
        if (logger.isDebugEnabled()) {
            logger.trace("Entering processWaitingForStateChangesState()");
        }

        if (!isNotifiedJobData(resource)) {
            return;
        }

        StateEnumeration state = resource.getState();

        JobIdStatePairType notifiedJobData = removeNotifiedJobData(resource);
        String localJobId = notifiedJobData.getId();
        StateEnumeration localJobState = notifiedJobData.getState();

        //ignore redundant local job state notifications
        if (localJobState.equals(state))
        {
            if (logger.isDebugEnabled()) {
               logger.trace("Ignoring redendant local job state notifications");
            }
            return;
        }

        String[] localJobIds = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);

        int i = 0;
        boolean jobIdFound = false;
        if (localJobIds != null) {
            for (i = 0; i < localJobIds.length; i++) {
                if (localJobId.equals(localJobIds[i])) {
                    jobIdFound = true;
                    break;
                }
            }
        }

        if (!jobIdFound)
        {
            String errorMessage =
                i18n.getMessage(Resources.JOB_ID_NOT_FOUND, localJobId);
            logger.error(errorMessage);

            return;
        }

        if (logger.isDebugEnabled()) {
            logger.debug("new job state for " + localJobId + " is "
                    + localJobState.toString());
        }

        ((ManagedExecutableJobResourceDataType)resource.getResourceData())
        .setLocalJobState(i, localJobState);

        if (state.equals(StateEnumeration.Suspended))
        {
            if (allJobProcessesHaveResumed(resource))
            {
                processNewState(
                    resource, InternalStateEnumeration.Resume, true);
            }
        }

        if (localJobState.equals(StateEnumeration.Pending))
        {
            try
            {
                resource.setState(localJobState);
            }
            catch (Exception e)
            {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }
        } else
        if (localJobState.equals(StateEnumeration.Active))
        {
            if (allJobProcessesAreActive(resource))
            {
                processNewState(
                    resource, InternalStateEnumeration.OpenStdout, true);
            }

        } else
        if (localJobState.equals(StateEnumeration.Suspended))
        {
            try {
                resource.setState(localJobState);
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }

            processNewState(
                resource, InternalStateEnumeration.Suspend, true);
        } else
        if (localJobState.equals(StateEnumeration.Done))
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Processing Done local job state.");
            }
            //make sure all processes are done before finishing up
            if (allJobProcessesAreDone(resource))
            {
                if (logger.isDebugEnabled()) {
                    logger.debug("All job processes are Done.");
                }

                processNewState(
                    resource, InternalStateEnumeration.MergeStdout, true);
            }
        } else
        if (localJobState.equals(StateEnumeration.Failed))
        {
            Integer exitCode = (Integer) resource.getResourceProperty(
                ManagedExecutableJobConstants.RP_EXIT_CODE);
            resource.setFault(
                createFaultFromErrorCode(
                    resource,
                    ScriptCommandEnumeration.submit,
                    exitCode.intValue(),
                    Resources.ABNORMAL_PROCESS_TERMINATION));
            systemCancel(resource);
        }

        if (logger.isDebugEnabled()) {
            logger.trace("Leaving processWaitingForStateChangesState()");
        }

        performanceLogger.stop("processWaitingForStateChangesState");
    }

    private static void processMergeStdoutState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processMergeStdoutState

        JobManagerScript script = null;

        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        PositiveInteger countObj = jobDescription.getCount();
        int count = countObj.intValue();

        String localResourceManager = (String) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER);
        if (count > 1 && (
            localResourceManager.equals(
                ManagedJobFactoryConstants.FACTORY_TYPE.FORK) ||
            localResourceManager.equals(
                ManagedJobFactoryConstants.FACTORY_TYPE.CONDOR))) {
            try
            {
                script = runScript(resource, ScriptCommandEnumeration.poll);
            }
            catch (Exception e)
            {
                if (script != null) {
                    resource.setFault(
                        createFaultFromErrorCode(
                            resource,
                            script.getCommand(),
                            0,
                            script.getFailureMessage()));
                }
                else
                {
                    String errorMessage = i18n.getMessage(
                        Resources.SCRIPT_EXECUTION_ERROR);
                    FaultType fault = FaultUtils.makeFault(
                        FaultType.class,
                        (ResourceKey) resource.getID(),
                        errorMessage,
                        e,
                        resource.getState(),
                        null,
                        0);
                    resource.setFault(fault);
                }

                systemCancel(resource);
            }

            //wait for script to finish
            synchronized (script) {
                while (!script.isDone())
                {
                    try {
                        script.wait(2000);
                    } catch (Exception e) {
                        String errorMessage =
                            i18n.getMessage(
                                Resources.INTERRUPTED_JOB_SCRIPT_WAIT,
                                script.getCommand());
                        logger.error(errorMessage, e);
                    }
                }
            }
            if (logger.isDebugEnabled()) {
                logger.trace("Done waiting for " + script.getCommand()
                            + " script");
            }

            int errorCode = script.getError();
            if (errorCode != 0) {
                resource.setFault(
                    createFaultFromErrorCode(
                        resource,
                        script.getCommand(),
                        errorCode,
                        script.getFailureMessage()));

                processNewState(
                    resource, InternalStateEnumeration.SystemCancel, true);

                performanceLogger.stop("processMergeStdoutState");

                return;
            }
        }

        Object fileStageOut = jobDescription.getFileStageOut();
        Object holdState = jobDescription.getHoldState();
        if (   (fileStageOut != null)
            || (   (holdState != null)
                && (holdState.equals(StateEnumeration.StageOut))))
        {
            logger.debug(getLoggingPrefix(resource) +
                "Not skipping StageOut");
            processNewState(resource, InternalStateEnumeration.StageOut, true);
        }
        else
        {
            logger.debug(getLoggingPrefix(resource) +
                "Skipping StageOut and going straight to CleanUp");
            processNewState(resource, InternalStateEnumeration.CleanUp, true);
        }

        performanceLogger.stop("processMergeStdoutState");
    }

    private static void processStageOutHoldState(
        ManagedExecutableJobResource        resource)
    {
        boolean holding = ((Boolean) resource.getResourceProperty(
            ManagedJobConstants.RP_HOLDING)).booleanValue();
        if (!holding) {
            processNewState(resource, InternalStateEnumeration.StageOut, true);
        } else {
            JobDescriptionType jobDescription
                = ((ServiceLevelAgreementType) resource.getResourceProperty(
                    ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
            if (jobDescription.getFileStageOut() == null) {
                resource.setFault(
                    createStagingHoldFault(
                        resource,
                        StagingTypeEnumeration.StageOut));
                resource.setResourceProperty(
                    ManagedJobConstants.RP_HOLDING,
                    false);

                systemCancel(resource);
            } else {
                try {
                    resource.setState(StateEnumeration.StageOut);
                } catch (Exception e) {
                    String errorMessage = i18n.getMessage(
                        Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                    throw new RuntimeException(errorMessage, e);
                }
            }
        }
    }

    private static void processStageOutState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processStageOutState

        StateEnumeration holdState
            = (StateEnumeration) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.StageOut))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.StageOutHold, true);

            performanceLogger.stop("processStageOutState");

            return;
        }

        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        TransferRequestType fileStageOut = jobDescription.getFileStageOut();

        boolean staging = false;
        if (fileStageOut != null)
        {
            staging = true;

            try {
                resource.setState(StateEnumeration.StageOut);
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }

            //submit a transfer request to RFT
            try {
                EndpointReferenceType stageOutEndpoint
                    = stagingHelper.submitStagingRequest(resource,fileStageOut);
                resource.setResourceDatum(
                    ManagedExecutableJobConstants.RD_STAGE_OUT_ENDPOINT,
                    stageOutEndpoint);
            }
            catch (Exception e)
            {
                resource.setFault(
                    createStagingFault(
                        resource,
                        StagingTypeEnumeration.StageOut,
                        e));
                systemCancel(resource);

                performanceLogger.stop("processStageOutState");

                return;
            }
        }

        if (staging)
        {
            processNewState(
                resource, InternalStateEnumeration.StageOutResponse, false);
        }
        else
        {
            processNewState(
                resource, InternalStateEnumeration.CleanUp, true);
        }

        performanceLogger.stop("processStageOutState");
    }

    private static void processStageOutResponseState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processStageOutResponseState()");
        }

        String source = null;
        String destination = null;
        String attribute = "fileStageOut";

        EndpointReferenceType transferJobEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_STAGE_OUT_ENDPOINT);
        if (transferJobEndpoint == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn("Missing transfer job endpoint. Continuing...");
            }

            processNewState(
                resource, InternalStateEnumeration.CleanUp, true);

            return;
        }

        RequestStatusType status
            = StagingListener.getNextStatus(transferJobEndpoint);
        if (status == null)
        {
            status = stagingHelper.poll(resource,transferJobEndpoint);
            if (status == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("Missing RFT status. "
                                + "Probably an out-of-order notification.");
                }
                return;
            }
        }

        //handle failed fileStageOut, but leave systemCancel() for later
        if (   status.getRequestStatus().equals(
                RequestStatusTypeEnumeration.Failed)
            && (resource.getFault() == null))
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Detected a StageOut transfer failure");
            }

            BaseFaultType transferFault
                = stagingHelper.getTransferFaultFromRP(status.getFault());

            //set fault RP
            FaultType fault = null;
            if (transferFault != null) {
                fault = FaultUtils.createStagingFault(
                    source,
                    destination,
                    attribute,
                    resource.getState(),
                    null,
                    (ResourceKey) resource.getID(),
                    transferFault);
            } else {
                fault = FaultUtils.createStagingFault(
                    source,
                    destination,
                    attribute,
                    resource.getState(),
                    null,
                    (ResourceKey) resource.getID(),
                    (Exception) new Exception().fillInStackTrace());
            }

            resource.setFault(fault);
        }

        if (!processCompletedStagingJob(
            resource, status, InternalStateEnumeration.CleanUp,
            transferJobEndpoint))
        {
            systemCancel(resource);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("leaving processStageOutResponseState()");
        }
    }

    private static void processCleanUpHoldState(
        ManagedExecutableJobResource        resource)
    {
        boolean holding = ((Boolean) resource.getResourceProperty(
            ManagedJobConstants.RP_HOLDING)).booleanValue();
        if (!holding) {
            processNewState(
                resource, InternalStateEnumeration.CleanUp, true);
        } else {
            try {
                if (logger.isDebugEnabled()) {
                    logger.trace("Notifying of CleanUp state");
                }
                resource.setState(StateEnumeration.CleanUp);
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    private static void processCleanUpState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processCleanUpState

        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.CleanUp))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.CleanUpHold, true);

            performanceLogger.stop("processCleanUpState");

            return;
        }

        try {
            if (logger.isDebugEnabled()) {
                logger.trace("Notifying of CleanUp state");
            }
            resource.setState(StateEnumeration.CleanUp);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }

        JobDescriptionType jobDescription = resource.getJobDescription();
        if (jobDescription.getFileCleanUp() != null)
        {
            logger.debug(getLoggingPrefix(resource) +
                "Not skipping FileCleanUp");
            processNewState(
                resource, InternalStateEnumeration.FileCleanUp, true);
        }
        else
        {
            logger.debug(getLoggingPrefix(resource) +
                "Skipping FileCleanUp and going straight to CacheCleanUp");
            processNewState(
                resource, InternalStateEnumeration.CacheCleanUp, true);
        }

        performanceLogger.stop("processCleanUpState");
    }

    private static void processUserCancelState(
        ManagedExecutableJobResource        resource)
    {
        kill(resource);

        JobDescriptionType jobDescription = resource.getJobDescription();
        if (jobDescription.getFileCleanUp() != null)
        {
            logger.debug(getLoggingPrefix(resource) +
                "Not skipping FailureFileCleanUp");
            processNewState(
                resource, InternalStateEnumeration.FailureFileCleanUp, true);
        }
        else
        {
            logger.debug(getLoggingPrefix(resource) +
                "Skipping FailureFileCleanUp and going straight to "
                + "FailureCacheCleanUp");
            processNewState(
                resource, InternalStateEnumeration.FailureCacheCleanUp, true);
        }
    }

    private static void processSystemCancelState(
        ManagedExecutableJobResource        resource)
    {
        kill(resource);

        processNewState(
            resource, InternalStateEnumeration.FailureFileCleanUp, true);
    }

    private static void processFileCleanUpState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processFileCleanUpState

        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        DeleteRequestType fileCleanUp = jobDescription.getFileCleanUp();

        if (fileCleanUp != null)
        {
            processNewState(
                resource, InternalStateEnumeration.FileCleanUpResponse, false);

            fileCleanUp(resource);
        }
        else
        {
            processNewState(
                resource, InternalStateEnumeration.CacheCleanUp, true);
        }

        performanceLogger.stop("processFileCleanUpState");
    }

    private static void processFileCleanUpResponseState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processFileCleanUpResponseState()");
        }

        EndpointReferenceType transferJobEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CLEAN_UP_ENDPOINT);
        if (transferJobEndpoint == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn("Missing transfer job endpoint. Continuing...");
            }

            processNewState(
                resource, InternalStateEnumeration.CacheCleanUp, true);

            return;
        }

        RequestStatusType status
            = StagingListener.getNextStatus(transferJobEndpoint);
        if (status == null)
        {
            status = stagingHelper.poll(resource,transferJobEndpoint);
            if (status == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("Missing RFT status. "
                                + "Probably an out-of-order notification.");
                }
                return;
            }
        }

        if (!processCompletedStagingJob(
            resource, status, InternalStateEnumeration.CacheCleanUp,
            transferJobEndpoint))
        {
            //go directly to FailureCacheCleanUp since systemCancel() will
            //repeat file clean up as FailureFileCleanUp
            processNewState(
                resource,
                InternalStateEnumeration.FailureCacheCleanUp,
                true);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("leaving processFileCleanUpResponseState()");
        }
    }

    private static void processFailureFileCleanUpState(
        ManagedExecutableJobResource        resource)
    {
        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        DeleteRequestType fileCleanUp = jobDescription.getFileCleanUp();

        if (fileCleanUp != null)
        {
            processNewState(
                resource,
                InternalStateEnumeration.FailureFileCleanUpResponse,
                false);

            fileCleanUp(resource);
        }
        else
        {
            processNewState(
                resource, InternalStateEnumeration.FailureCacheCleanUp, true);
        }
    }

    private static void processFailureFileCleanUpResponseState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processFailureFileCleanUpResponseState()");
        }

        EndpointReferenceType transferJobEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CLEAN_UP_ENDPOINT);
        if (transferJobEndpoint == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.warn("Missing transfer job endpoint. Continuing...");
            }

            processNewState(
                resource, InternalStateEnumeration.FailureCacheCleanUp, true);

            return;
        }

        RequestStatusType status
            = StagingListener.getNextStatus(transferJobEndpoint);
        if (status == null)
        {
            status = stagingHelper.poll(resource,transferJobEndpoint);
            if (status == null)
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("Missing RFT status. "
                                + "Probably an out-of-order notification.");
                }
                return;
            }
        }

        if (!processCompletedStagingJob(
            resource, status, InternalStateEnumeration.FailureCacheCleanUp,
            transferJobEndpoint))
        {
            //go directly to FailureCacheCleanUp since systemCancel() will
            //cause endless looping of FailureFileCleanUp
            processNewState(
                resource,
                InternalStateEnumeration.FailureCacheCleanUp,
                true);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("leaving processFailureFileCleanUpResponseState()");
        }
    }

    private static void processCacheCleanUpState(
        ManagedExecutableJobResource        resource)
    {
        performanceLogger.start(); //processCacheCleanUpState

        boolean success = cacheCleanUp(resource);
        if (success) {
            processNewState(resource, InternalStateEnumeration.Done, true);
        }

        performanceLogger.stop("processCacheCleanUpState");
    }

    private static void processFailureCacheCleanUpState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled())
        {
            logger.trace("Entering processFailureCacheCleanUpState()");
        }

        cacheCleanUp(resource);

        processNewState(resource, InternalStateEnumeration.Failed, true);

        if (logger.isDebugEnabled())
        {
            logger.trace("Leaving processFailureCacheCleanUpState()");
        }
    }

    private static void sendUsagePacket(
        ManagedExecutableJobResource        resource)
    {
        JobDescriptionType jobDescription = resource.getJobDescription();

        GramUsageMonitorPacket usagePacket = new GramUsageMonitorPacket();

        usagePacket.setTimestamp(System.currentTimeMillis());
        usagePacket.setCreationTime((Date)resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_CREATION_TIME));
        usagePacket.setLocalResourceManager((String)resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER));
        usagePacket.setJobCredentialEndpointUsed(
            jobDescription.getJobCredentialEndpoint()!=null?true:false);
        usagePacket.setFileStageInUsed(
            jobDescription.getFileStageIn()!=null?true:false);
        usagePacket.setFileStageOutUsed(
            jobDescription.getFileStageOut()!=null?true:false);
        usagePacket.setFileCleanUpUsed(
            jobDescription.getFileCleanUp()!=null?true:false);
        StateEnumeration holdState = jobDescription.getHoldState();
        if (holdState != null)
        {
            usagePacket.setCleanUpHoldUsed(
                holdState.equals(StateEnumeration.CleanUp)?true:false);
        } else
        {
            usagePacket.setCleanUpHoldUsed(false);
        }
        JobTypeEnumeration jobType = jobDescription.getJobType();
        if (jobType == null)
        {
            jobType = JobTypeEnumeration.multiple;
        }
        usagePacket.setJobType(jobType);

        StateEnumeration state = resource.getState();
        if (state.equals(StateEnumeration.Failed))
        {
            FaultType fault = (FaultType)resource.getFault();
            if (fault != null)
            {
                usagePacket.setGt2ErrorCode(fault.getGt2ErrorCode());
                usagePacket.setFaultClass(fault.getClass());
            }
            else
            {
                logger.error(i18n.getMessage(
                    Resources.NO_FAULT_ON_FAILURE,
                    ((ResourceKey) resource.getID()).toString()));
            }
        }

        if (usageConfig != null)
        {
            List targets = usageConfig.getTargets();
            if (targets != null)
            {
                usagePacket.setHostIP(usageConfig.getContainerHost());
                usagePacket.sendPacket(targets);
            }
        }

    }

    private static void processDoneHoldState(
        ManagedExecutableJobResource        resource)
    {
        StateEnumeration state = resource.getState();
        resource.setFault(
            FaultUtils.createUnsupportedFeatureFault(
                "Done", "holdState", state,
                null, -1, (ResourceKey) resource.getID()));
        resource.setResourceProperty(
            ManagedJobConstants.RP_HOLDING,
            false);

        processNewState(resource, InternalStateEnumeration.Failed, true);
    }

    private static void processDoneState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled())
        {
            logger.trace("Entering processDoneState()");
        }
        performanceLogger.start(); //processDoneState

        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.Done))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.DoneHold, true);

            performanceLogger.stop("processDoneState");

            if (logger.isDebugEnabled())
            {
                logger.trace("Leaving processDoneState() for hold");
            }
            return;
        }

        try {
            resource.setState(StateEnumeration.Done);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }

        logAuditRecord(false, resource, true, true);
        unsubscribeAtJobStateMonitor(resource);
        logJobSucceeded(resource);

        sendUsagePacket(resource);

        performanceLogger.stop("processDoneState");

        if (logger.isDebugEnabled())
        {
            logger.trace("Leaving processDoneState()");
        }
    }

    private static void processFailedHoldState(
        ManagedExecutableJobResource        resource)
    {
        StateEnumeration state = resource.getState();
        resource.setFault(
            FaultUtils.createUnsupportedFeatureFault(
                "Failed", "holdState", state,
                null, -1, (ResourceKey) resource.getID()));
        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_HOLD_STATE,
            null);
        resource.setResourceProperty(
            ManagedJobConstants.RP_HOLDING,
            false);

        processNewState(resource, InternalStateEnumeration.Failed, true);
    }

    private static void processFailedState(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled())
        {
            logger.trace("Entering processFailedState()");
        }

        StateEnumeration holdState
            = (StateEnumeration)resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.Failed))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.FailedHold, true);

            return;
        }

        logger.debug("fault:\n", resource.getFault());

        try {
            resource.setState(StateEnumeration.Failed);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }

        logAuditRecord(false, resource, false, true);
        unsubscribeAtJobStateMonitor(resource);
        logJobFailed(resource);

        sendUsagePacket(resource);

        if (logger.isDebugEnabled())
        {
            logger.trace("Leaving processFailedState()");
        }
    }

    private static void processSuspendedHoldState(
        ManagedExecutableJobResource        resource)
    {
        StateEnumeration state = resource.getState();
        resource.setFault(
            FaultUtils.createUnsupportedFeatureFault(
                "Suspended", "holdState", state,
                null, -1, (ResourceKey) resource.getID()));
        resource.setResourceProperty(
            ManagedJobConstants.RP_HOLDING,
            false);

        //processNewState(resource, InternalStateEnumeration.SystemCancel, true);
        systemCancel(resource);
    }

    private static void processSuspendState(
        ManagedExecutableJobResource        resource)
    {
        StateEnumeration holdState = (StateEnumeration)resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_HOLD_STATE);
        if ((holdState != null) && holdState.equals(StateEnumeration.Suspended))
        {
            resource.setResourceProperty(
                ManagedJobConstants.RP_HOLDING,
                true);

            processNewState(
                resource, InternalStateEnumeration.SuspendedHold, true);
            return;
        }

        InternalStateEnumeration internalState
            = resource.getInternalState();

        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_SUSPENDED_INTERNAL_STATE,
            internalState);

        processNewState(
            resource,
            InternalStateEnumeration.WaitingForStateChanges,
            false);
    }

    private static void processResumeState(
        ManagedExecutableJobResource        resource)
    {

        try {
            resource.setState(StateEnumeration.Suspended);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.RESOURCE_SET_STATE_NOTIFY_ERROR);
            throw new RuntimeException(errorMessage, e);
        }

        InternalStateEnumeration suspendedInternalState
            = (InternalStateEnumeration) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_SUSPENDED_INTERNAL_STATE);

        processNewState(resource, suspendedInternalState, false);
    }
    //------------------------------------------------//


    //-------------- Helper Methods ------------------//
    private static void processNewState(
        ManagedExecutableJobResource        resource,
        InternalStateEnumeration            newInternalState,
        boolean                             addToRunQueue)
    {
        try
        {
            resource.setInternalState(newInternalState);
        }
        catch (ResourceException re)
        {
            throw new RuntimeException(re);
        }

        if (logger.isDebugEnabled())
        {
            logger.debug(getLoggingPrefix(resource) +
                "In internal state " + newInternalState);
        }

        if (addToRunQueue)
        {
            if (logger.isDebugEnabled()) {
                logger.trace("Adding to RunQueue.");
            }
            RunQueue.getInstance().add(
                (ResourceKey)resource.getID());
        }
    }

    private static boolean allJobProcessesAreActive(
        ManagedExecutableJobResource        resource)
    {
        ResourcePropertySet             props =
                resource.getResourcePropertySet();
        StateEnumeration[] localJobStates =
            (StateEnumeration[])resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_JOB_STATES);
        int size = localJobStates.length;
        int active = 0;

        if (localJobStates != null) {
            for (int i = 0; i < size; i++) {
                if (localJobStates[i].equals(StateEnumeration.Active)) {
                    active++;
                }
            }

            if (active == size) {
                return true;
            }
        }

        return false;
    }

    private static boolean allJobProcessesAreDone(
        ManagedExecutableJobResource        resource)
    {
        ResourcePropertySet             props =
                resource.getResourcePropertySet();
        StateEnumeration[] localJobStates =
            (StateEnumeration[])resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_JOB_STATES);
        int size = localJobStates.length;
        int done = 0;

        if (localJobStates != null) {
            for (int i = 0; i < size; i++) {
                if (localJobStates[i].equals(StateEnumeration.Done)) {
                    done++;
                }
            }

            if (done == size) {
                return true;
            }
        }

        return false;
    }

    private static boolean allJobProcessesHaveResumed(
        ManagedExecutableJobResource        resource)
    {
        ResourcePropertySet             props =
                resource.getResourcePropertySet();
        StateEnumeration[] localJobStates =
            (StateEnumeration[])resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_JOB_STATES);
        int size = localJobStates.length;
        int suspended = 0;

        if (localJobStates != null) {
            for (int i = 0; i < size; i++) {
                if (localJobStates[i].equals(StateEnumeration.Suspended)) {
                    suspended++;
                }
            }

            if (suspended != size) {
                return true;
            }
        }

        return false;
    }

    protected static void insertNotifiedJobData(
        ManagedExecutableJobResource        resource,
        String                              jobId,
        StateEnumeration                    state)
    {
        if (logger.isDebugEnabled())
        {
            logger.debug(getLoggingPrefix(resource) +
                 jobId + " " + state + " " + resource.getID());
        }

        JobIdStatePairType[] notifiedJobData
            = (JobIdStatePairType[]) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_NOTIFIED_JOB_DATA);

        List newNotifiedJobData;

        if (notifiedJobData == null) {
            newNotifiedJobData = new ArrayList();
        }
        else {
            newNotifiedJobData = new ArrayList(Arrays.asList(notifiedJobData));
        }
        if (logger.isDebugEnabled()) {
            logger.trace("newNotifiedJobData is of type "
                        + newNotifiedJobData.getClass().getName());
        }

        newNotifiedJobData.add(new JobIdStatePairType(
            jobId, state));

        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_NOTIFIED_JOB_DATA,
            (JobIdStatePairType[])newNotifiedJobData.toArray(
                new JobIdStatePairType[0]));
    }

    protected static JobIdStatePairType removeNotifiedJobData(
        ManagedExecutableJobResource        resource)
    {
        JobIdStatePairType[] notifiedJobData
            = (JobIdStatePairType[]) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_NOTIFIED_JOB_DATA);

        List newNotifiedJobData = new ArrayList(Arrays.asList(notifiedJobData));

        newNotifiedJobData
            = newNotifiedJobData.subList(1, newNotifiedJobData.size());

        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_NOTIFIED_JOB_DATA,
            (JobIdStatePairType[])newNotifiedJobData.toArray(
                new JobIdStatePairType[0]));

        return notifiedJobData[0];
    }

    private static void kill(
        ManagedExecutableJobResource        resource)
    {
        //run cancel job script command
        JobManagerScript script = null;
        try {
            script = runScript(
                resource,
                ScriptCommandEnumeration.cancel);
        } catch (Exception e) {
            if (script != null) {
                resource.setFault(
                    createFaultFromErrorCode(
                        resource,
                        script.getCommand(),
                        0,
                        script.getFailureMessage()));
            } else {
                String errorMessage = i18n.getMessage(
                    Resources.SCRIPT_EXECUTION_ERROR);
                FaultType fault = FaultUtils.makeFault(
                    FaultType.class,
                    (ResourceKey) resource.getID(),
                    errorMessage,
                    e,
                    resource.getState(),
                    null,
                    0);
                resource.setFault(fault);
            }
        }
    }

    private static boolean processCompletedStagingJob(
        ManagedExecutableJobResource        resource,
        RequestStatusType                   status,
        InternalStateEnumeration            nextState,
        EndpointReferenceType               transferJobEndpoint)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processCompletedStagingJob()");
        }

        RequestStatusTypeEnumeration requestState
            = status.getRequestStatus();
        BaseFaultType transferFault
            = stagingHelper.getTransferFaultFromRP(status.getFault());

        //staging finished
        if (requestState.equals(RequestStatusTypeEnumeration.Done))
        {
            //all transfers finished
            if (logger.isDebugEnabled()) {
                logger.debug("Transfer finished successufully.");
            }

            try {
                stagingHelper.destroyTransferJob(
                      resource, transferJobEndpoint);
            } catch(Exception e) {
                String errorMessage = i18n.getMessage(
                        Resources.RFT_RESOURCE_DESTROY_ERROR);
                logger.error(errorMessage, e);
            }

            StateMachine.processNewState(resource, nextState, true);
        }
        else
        if (requestState.equals(RequestStatusTypeEnumeration.Failed))
        {
            //process failed stage
            if (logger.isDebugEnabled())
            {
                logger.debug("Handling non-StageIn failures");
            }

            try
            {
                stagingHelper.destroyTransferJob(
                      resource, transferJobEndpoint);
            }
            catch(Exception e)
            {
                String errorMessage = i18n.getMessage(
                        Resources.RFT_RESOURCE_DESTROY_ERROR);
                logger.error(errorMessage, e);
            }

            return false;
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving processCompletedStagingJob()");
        }

        return true;
    }

    private static boolean fileCleanUp(
        ManagedExecutableJobResource        resource)
    {
        JobDescriptionType jobDescription
            = ((ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT)).getJob();
        DeleteRequestType fileCleanUp = jobDescription.getFileCleanUp();

        if (fileCleanUp != null)
        {
            //submit a delete request to RFT
            try {
                EndpointReferenceType cleanUpEndpoint
                    = stagingHelper.submitStagingRequest(resource, fileCleanUp);
                resource.setResourceDatum(
                    ManagedExecutableJobConstants.RD_CLEAN_UP_ENDPOINT,
                    cleanUpEndpoint);
            }
            catch (Exception e)
            {
                if (resource.getFault() == null)
                {
                    resource.setFault(
                        createStagingFault(
                            resource,
                            StagingTypeEnumeration.CleanUp,
                            e));
                }
                else
                {
                    logger.error(i18n.getMessage(Resources.SECONDARY_FAULT), e);
                }

                processNewState(
                    resource,
                    InternalStateEnumeration.FailureCacheCleanUp,
                    true);

                return false;
            }
        } else
        {
            //no delete request, so don't rely on RFT notifications to continue
            RunQueue.getInstance().add(
                (ResourceKey)resource.getID());
        }

        return true;
    }

    private static boolean cacheCleanUp(
        ManagedExecutableJobResource        resource)
    {
        //run cache clean up script
        JobManagerScript script = null;
        try {
            script = runScript(
                resource,
                ScriptCommandEnumeration.cache_cleanup);
        } catch (Exception e) {
            if (script != null) {
                if (resource.getFault() != null)
                {
                    resource.setFault(
                        createFaultFromErrorCode(
                            resource,
                            script.getCommand(),
                            0,
                            script.getFailureMessage()));
                }
            } else {
                String errorMessage = i18n.getMessage(
                    Resources.SCRIPT_EXECUTION_ERROR);
                FaultType fault = FaultUtils.makeFault(
                    FaultType.class,
                    (ResourceKey) resource.getID(),
                    errorMessage,
                    e,
                    resource.getState(),
                    null,
                    0);
                resource.setFault(fault);
            }

            StateMachine.systemCancel(resource);

            return false;
        }

        //wait for script to finish
        synchronized (script) {
            while (!script.isDone())
            {
                try {
                    script.wait(2000);
                } catch (Exception e) {
                    String errorMessage =
                        i18n.getMessage(Resources.INTERRUPTED_JOB_SCRIPT_WAIT,
                                        script.getCommand());
                    logger.error(errorMessage, e);

                    return false;
                }
            }
        }
        if (logger.isDebugEnabled()) {
            logger.trace("Done waiting for " + script.getCommand()
                        + " script");
        }

        //get error code
        int errorCode = script.getError();
        if (logger.isDebugEnabled()) {
            logger.debug("script return code: " + errorCode);
        }

        //handle error code
        if (errorCode != 0)
        {
            logger.debug("script return code means error!");
            boolean canceled = ((Boolean) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CANCELED))
                .booleanValue();
            if (!canceled && (resource.getFault() != null))
            {
                resource.setFault(
                    createFaultFromErrorCode(
                        resource,
                        script.getCommand(),
                        errorCode,
                        script.getFailureMessage()));
            }

            processNewState(resource, InternalStateEnumeration.Failed, true);
            return false;
        }

        return true;
    }

    /**
     * Run a job manager script to interface with the local scheduling
     * system.
     *
     * @param scriptCommand Name of the Perl method to invoke.
     *
     * @throws IOException Error forking process or creating script output
     *                     reader.
     * @see org.globus.exec.service.job.jobmanager.JobManagerScript
     */
    protected static JobManagerScript runScript(
        ManagedExecutableJobResource        resource,
        ScriptCommandEnumeration            scriptCommand)
        throws                              IOException
    {
        String globusLocation = null;
        //try {
            globusLocation
                = ManagedJobFactoryHome.getConfiguration().getGlobusLocation();
        //} catch (Exception e) {
        //    logger.error(
        //       "unable to obtain globusLocation from factory configuration",
        //       e);
        //}

        String perlAttributesString =
            resource.getPerlJobDescription().toPerlString();
        if (logger.isDebugEnabled()) {
            logger.debug("Perl Job Description: " + perlAttributesString);
        }

        JobManagerScript newScript = new JobManagerScript(
            (String) resource.getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID),
            globusLocation,
            (String) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER),
            scriptCommand,
            perlAttributesString,
            resource.getScriptEnvironment());
            //resource);
        resource.setScript(newScript);
        if (logger.isTraceEnabled()) {
            logger.trace("running script " + newScript.getCommand());
        }
        newScript.start();

        return newScript;
    }

    protected static void setLocalJobId(
        ManagedExecutableJobResource        resource,
        String                              localJobId,
        StateEnumeration                    state)
    {
        StringTokenizer tokenizer = new StringTokenizer(localJobId, ",");
        int tokenCount = tokenizer.countTokens();

        String[] localJobIds = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
        String[] localJobStates = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_STATES);

        List newLocalJobIds;
        List newLocalJobStates;

        if (localJobIds == null) {
            newLocalJobIds = new ArrayList();
        }
        else {
            newLocalJobIds = new ArrayList(Arrays.asList(localJobIds));
        }
        if (logger.isDebugEnabled()) {
            logger.trace("newLocalJobIds is of type "
                        + newLocalJobIds.getClass().getName());
        }

        if (localJobStates == null) {
            newLocalJobStates = new ArrayList();
        }
        else {
            newLocalJobStates = new ArrayList(Arrays.asList(localJobStates));
        }

        for (int i = 0; i < tokenCount; i++) {
            String token = tokenizer.nextToken();
            logger.debug("Saving local job ID " + token);

            newLocalJobIds.add(token);
            newLocalJobStates.add(state);
        }

        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS,
            (String[])newLocalJobIds.toArray(new String[0]));
        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_STATES,
            (StateEnumeration[])newLocalJobStates.toArray(
                            new StateEnumeration[0]));
    }

    protected static boolean isNotifiedJobData(
        ManagedExecutableJobResource        resource)
    {
        JobIdStatePairType[] notifiedJobData
            = (JobIdStatePairType[]) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_NOTIFIED_JOB_DATA);

        if ((notifiedJobData != null) && (notifiedJobData.length > 0))
        {
            return  true;
        }

        return false;
    }

    private static FaultType createStagingHoldFault(
        ManagedExecutableJobResource        resource,
        StagingTypeEnumeration              stagingType)
    {

        String attribute = null;
        if (stagingType.equals(StagingTypeEnumeration.StageIn)) {
            attribute = "fileStageIn";
        } else if (stagingType.equals(StagingTypeEnumeration.StageOut)) {
            attribute = "fileStageOut";
        } else {
            attribute = "fileCleanUp";
        }

        Exception cause = new Exception(
            i18n.getMessage("badHoldRequest",
                new String[] { stagingType.toString() }));
        cause.fillInStackTrace();

        return FaultUtils.createStagingFault(
            null, null, attribute,
            resource.getState(),
            stagingType.toString(),
            (ResourceKey) resource.getID(), cause);
    }

    private static FaultType createStagingFault(
        ManagedExecutableJobResource        resource,
        StagingTypeEnumeration              stagingType,
        Exception                           cause)
    {

        String attribute = null;
        if (stagingType.equals(StagingTypeEnumeration.StageIn)) {
            attribute = "fileStageIn";
        } else if (stagingType.equals(StagingTypeEnumeration.StageOut)) {
            attribute = "fileStageOut";
        } else {
            attribute = "fileCleanUp";
        }

        return FaultUtils.createStagingFault(
            null, null, attribute,
            resource.getState(),
            stagingType.toString(),
            (ResourceKey) resource.getID(), cause);
    }

    private static FaultType createFaultFromErrorCode(
        ManagedExecutableJobResource        resource,
        ScriptCommandEnumeration            scriptCommand,
        int                                 errorCode)
    {
        return createFaultFromErrorCode(
            resource, scriptCommand, errorCode, null);
    }

    private static FaultType createFaultFromErrorCode(
        ManagedExecutableJobResource        resource,
        ScriptCommandEnumeration            scriptCommand,
        int                                 errorCode,
        String                              failureMessage)
    {
        Exception cause = null;
        if (failureMessage != null)
        {
            cause = new Exception(failureMessage);
            cause.fillInStackTrace();
        }

        JobDescriptionType jobDescription = resource.getJobDescription();
        logger.trace("Creating fault from error code "
                    + Integer.toString(errorCode));

        switch(errorCode) {
            case GRAMProtocolErrorConstants.BAD_SCRIPT_ARG_FILE:
            {
                cause = new Exception(
                    i18n.getMessage(Resources.BAD_SCRIPT_ARG_FILE));
                cause.fillInStackTrace();

                return FaultUtils.createInternalFault(
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);
            }

            case GRAMProtocolErrorConstants.BAD_DIRECTORY:
                return FaultUtils.createInvalidPathFault(
                        jobDescription.getDirectory(), "directory",
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case 136/*INVALID_SCRATCH*/:
                String manager = ((ManagedExecutableJobResourceDataType)resource
                    .getResourceData()).getLocalResourceManager();
                ManagedJobFactoryResource factoryResource = null;
                try {
                    ManagedJobFactoryHome home
                        = ManagedJobFactoryHome.getHome();
                    factoryResource = (ManagedJobFactoryResource) home.find(
                        ManagedJobHelper.getResourceKey(manager));
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
                String scratchDir
                    = (String) factoryResource.getResourcePropertySet().get(
                        ManagedJobFactoryConstants.RP_SCRATCH_BASE_DIRECTORY)
                        .get(0);

                return FaultUtils.createInvalidPathFault(
                        scratchDir,
                        "scratchDir",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.EXECUTABLE_NOT_FOUND:
                return FaultUtils.createInvalidPathFault(
                        jobDescription.getExecutable(), "executable",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.STDIN_NOT_FOUND:
                return FaultUtils.createInvalidPathFault(
                        jobDescription.getStdin(), "stdin",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.INVALID_COUNT:
                return FaultUtils.createServiceLevelAgreementFault(
                        "count",
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.JOB_EXECUTION_FAILED:
                return FaultUtils.createExecutionFailedFault(
                        "stdin",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.INVALID_JOBTYPE:
            case 26 /*JOBTYPE_NOT_SUPPORTED*/:
                JobTypeEnumeration jobType = jobDescription.getJobType();
                if (jobType == null) {
                    jobType = JobTypeEnumeration.single;
                }

                return FaultUtils.createUnsupportedFeatureFault(
                        jobType.toString(), "jobType",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.EXECUTABLE_PERMISSIONS:
                return FaultUtils.createFilePermissionsFault(
                        jobDescription.getExecutable(), "executable",
                        resource.getState(),
                        scriptCommand.toString(), errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.ERROR_OPENING_STDOUT:
                return FaultUtils.createFilePermissionsFault(
                        jobDescription.getStdout(),
                        "stdout",
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.ERROR_OPENING_STDERR:
                return FaultUtils.createFilePermissionsFault(
                        jobDescription.getStderr(),
                        "stderr",
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            case GRAMProtocolErrorConstants.INVALID_SCRIPT_REPLY:
                return FaultUtils.createInternalFault(
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);

            default:
                String warnMessage =
                    i18n.getMessage(Resources.UNKNOWN_GRAM_FAULT_CODE,
                                    Integer.toString(errorCode));
                logger.warn(warnMessage);
                logger.debug("Offending Script Command: "
                            + scriptCommand.toString());
                return FaultUtils.createFault(
                        resource.getState(),
                        scriptCommand.toString(),
                        errorCode,
                        (ResourceKey) resource.getID(),
                        cause);
        }
    }

    private static void logAuditRecord(
        boolean                             insertRecord,
        ManagedExecutableJobResource        resource,
        boolean                             succeeded,
        boolean                             finished)
    {
        StringBuffer auditRecord = new StringBuffer();
        //insert or update record?
        if (insertRecord) {
            auditRecord.append("insert: ");
        }
        else {
            auditRecord.append("update: ");
        }

        //job_grid_id
        try
        {
            FactoryServiceConfiguration serviceConfiguration
                = ManagedJobFactoryService.getServiceConfiguration();
            String jobServicePath
                = serviceConfiguration.getManagedExecutableJobServicePath();
            String serviceURL = ServiceHost.getBaseURL() + jobServicePath;
            EndpointReferenceType jobEndpoint
                = AddressingUtils.createEndpointReference(
                    serviceURL,
                    (ResourceKey) resource.getID());
            Class resourceKeyType
                = ManagedExecutableJobHome.getHome().getKeyTypeClass();
            String jobGridId = AuditUtil.eprToGridId(jobEndpoint);
            auditRecord.append(
                AuditDatabaseAppender.format(jobGridId)).append(",");
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }

        //local_job_id
        StringBuffer localJobIdsString = new StringBuffer();
        String[] localJobIds = (String []) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
        if (localJobIds != null)
        {
            for (int index=0; index<localJobIds.length; index++)
            {
                localJobIdsString.append(localJobIds[index]);
                if (index < (localJobIds.length-1))
                {
                    localJobIdsString.append(";");
                }
            }
        }
        else
        {
            localJobIdsString.append("NULL");
        }
        auditRecord.append(
            AuditDatabaseAppender.format(
                localJobIdsString.toString())).append(",");

        //subject_name
        auditRecord.append(
            AuditDatabaseAppender.format((String) resource.getResourceProperty(
                ManagedJobConstants.RP_USER_SUBJECT))).append(",");

        //username
        auditRecord.append(
            AuditDatabaseAppender.format((String) resource.getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID))).append(",");

        //idempotence_id
        String idempotenceId = (String) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_IDEMPOTENCE_ID);
        if (idempotenceId == null)
        {
            idempotenceId = "NULL";
        }
        auditRecord.append(
            AuditDatabaseAppender.format(idempotenceId)).append(",");

        //creation_time
        auditRecord.append(
            AuditDatabaseAppender.format(resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CREATION_TIME).toString()))
                .append(",");

        //queued_time
        Date queuedTime = (Date) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_QUEUED_TIME);
        String queuedTimeString
            = (queuedTime!=null?queuedTime.toString():"NULL");
        auditRecord.append(
            AuditDatabaseAppender.format(queuedTimeString)).append(",");

        //stage_in_grid_id
        EndpointReferenceType stageInEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_STAGE_IN_ENDPOINT);
        String stageInEndpointString = null;
        if (stageInEndpoint == null)
        {
            stageInEndpointString = "NULL";
        }
        else
        {
            try
            {
                Class resourceKeyType
                    = ManagedExecutableJobHome.getHome().getKeyTypeClass();
                stageInEndpointString = AuditUtil.eprToGridId(stageInEndpoint);
            }
            catch (Exception e)
            {
                throw new RuntimeException(e);
            }
        }
        auditRecord.append(
            AuditDatabaseAppender.format(stageInEndpointString)).append(",");

        //stage_out_grid_id
        EndpointReferenceType stageOutEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_STAGE_OUT_ENDPOINT);
        String stageOutEndpointString = null;
        if (stageOutEndpoint == null)
        {
            stageOutEndpointString = "NULL";
        }
        else
        {
            try
            {
                Class resourceKeyType
                    = ManagedExecutableJobHome.getHome().getKeyTypeClass();
                stageOutEndpointString
                    = AuditUtil.eprToGridId(stageOutEndpoint);
            }
            catch (Exception e)
            {
                throw new RuntimeException(e);
            }
        }
        auditRecord.append(
            AuditDatabaseAppender.format(stageOutEndpointString)).append(",");

        //clean_up_grid_id
        EndpointReferenceType cleanUpEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_CLEAN_UP_ENDPOINT);
        String cleanUpEndpointString = null;
        if (cleanUpEndpoint == null)
        {
            cleanUpEndpointString = "NULL";
        }
        else
        {
            try
            {
                Class resourceKeyType
                    = ManagedExecutableJobHome.getHome().getKeyTypeClass();
                cleanUpEndpointString = AuditUtil.eprToGridId(cleanUpEndpoint);
            }
            catch (Exception e)
            {
                throw new RuntimeException(e);
            }
        }
        auditRecord.append(
            AuditDatabaseAppender.format(cleanUpEndpointString)).append(",");

        //globus_toolkit_version
        AuditDatabaseConfiguration auditDatabaseConfiguration
            = ManagedExecutableJobHome.getAuditDatabaseConfiguration();
        String globusToolkitVersion
            = auditDatabaseConfiguration.getGlobusVersion();
        auditRecord.append(
            AuditDatabaseAppender.format(globusToolkitVersion)).append(",");

        //resource_manager_type
        auditRecord.append(
            AuditDatabaseAppender.format((String) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER)))
                .append(",");

        //job_description
        ServiceLevelAgreementType jobDescription
            = (ServiceLevelAgreementType) resource.getResourceProperty(
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT);
        Element jobDescriptionElement = null;
        try
        {
            jobDescriptionElement = ObjectSerializer.toElement(
                jobDescription,
                ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT);
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
        auditRecord.append(
            AuditDatabaseAppender.format(
                XmlUtils.toString(jobDescriptionElement))).append(",");

        //success_flag
        auditRecord.append(
            AuditDatabaseAppender.format(new Boolean(succeeded).toString())).
            append(",");

        //finished_flag
        auditRecord.append(
            AuditDatabaseAppender.format(new Boolean(finished).toString()));

        auditLogger.debug(auditRecord.toString());
    }

    public static String getLoggingPrefix(
        ManagedExecutableJobResource        resource)
    {
        return getLoggingPrefix((ResourceKey) resource.getID());
    }

    public static String getLoggingPrefix(ResourceKey resourceKey)
    {
        return new String("[resourceKey:" + resourceKey.getValue() + "] ");
    }

    private static void logJobAccepted(
        ManagedExecutableJobResource        resource)
        throws                              ResourceException
    {
        logger.info(
            "Job " + ((ResourceKey)resource.getID()).getValue()
            + " accepted for local user '"
            + (String) resource.getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID)
            + "'");
    }

    private static void logJobSubmitted(
        ManagedExecutableJobResource        resource,
        String                              localJobId)
    {
        logger.info(
            "Job " + ((ResourceKey)resource.getID()).getValue()
            + " submitted with local job ID '"
            + localJobId
            + "'");
    }
    
    private static void logJobSucceeded(
        ManagedExecutableJobResource        resource)
    {
        logger.info(
            "Job " + ((ResourceKey)resource.getID()).getValue()
            + " finished successfully");
    }

    private static void logJobFailed(
        ManagedExecutableJobResource        resource)
    {
        logger.info(
            "Job " + ((ResourceKey)resource.getID()).getValue() + " failed");
    }
    
    static void subscribeAtJobStateMonitor(
            ManagedExecutableJobResource        resource)
            throws                              AlreadyRegisteredException
    {
        logger.debug("subscribing resource for JSM notifications");
        String[] localJobIds = (String[]) resource.getResourceDatum(
        ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
 
        if (localJobIds != null) {
            int size = localJobIds.length;
            for (int i = 0; i < size; i++) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Registering job ID " + localJobIds[i]
                                 +" for JSM subscriber " + resource.getID());
                }
                resource.getJobStateMonitor().registerJobID(
                    localJobIds[i], (ResourceKey) resource.getID());
            }
        }

        if (logger.isDebugEnabled()) {
                logger.debug("Exiting subscribeForNotifications()");
        }
    }

    static void unsubscribeAtJobStateMonitor(
            ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("unsubscribing resource for JSM notifications");
        }

        String[] localJobIds = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);

        if (localJobIds != null) {
            int size = localJobIds.length;

            for (int i = 0; i < size; i++) {
                try {
                    if (logger.isDebugEnabled()) {
                        logger.debug(
                            "Unregistering job ID from JSM subscribers");
                    }

                    resource.getJobStateMonitor().unregisterJobID(
                            localJobIds[i]);
                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Job ID Unregistered.");
                    }
                } catch (Exception e) {
                    String errorMessage = i18n.getMessage(
                        Resources.JOB_ID_UNREGISTRATION_ERROR);
                    logger.debug(errorMessage, e); //debug not error...
                }
            }
        }
    }    
    
    //------------------------------------------------//
}
