/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URLDecoder;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.generated.FileSystemMapType;
import org.globus.exec.generated.FileSystemMapsType;
import org.globus.exec.generated.FileSystemPathMappingType;
import org.globus.exec.generated.FtpServerType;
import org.globus.exec.utils.Resources;
import org.globus.util.GlobusURL;
import org.globus.util.I18n;
import org.globus.wsrf.encoding.ObjectDeserializationContext;

import org.xml.sax.InputSource;

public class FileMapping
{
    static I18n i18n = I18n.getI18n(Resources.class.getName());
    private static Log logger =
            LogFactory.getLog(FileMapping.class.getName());
    private FileSystemMapType map;
    public static final String FILE_MAP_CONFIG_FILE
            = "globus_gram_fs_map_config.xml";
    private static final String URL_ENCODING = "UTF-8";

    public FileMapping(String fileName, String manager)
        throws                              FileMappingException
    {
        FileInputStream in = null;

        try
        {
            in = new FileInputStream(fileName);

            ObjectDeserializationContext deserializer
                = new ObjectDeserializationContext(new InputSource(in),
                        FileSystemMapsType.class);

            deserializer.parse();
            FileSystemMapsType maps = (FileSystemMapsType)
                    deserializer.getValue();

            FileSystemMapType map_array[] = maps.getMap();

            if (map_array == null)
            {
                String errorMessage = i18n.getMessage(
                    Resources.FILE_MAP_CONFIG_EMPTY,
                    fileName);
                throw new RuntimeException(errorMessage);
            }

            for (int i = 0; i < map_array.length; i++)
            {
                if (map_array[i].getScheduler().equals(manager))
                {
                    this.map =  map_array[i];
                    break;
                }
            }

        }
        catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                Resources.FILE_MAP_INIT_FAILED);
            throw new FileMappingException(errorMessage, e);
        }
        finally
        {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (IOException ioe) {
                logger.warn(ioe);
            }
        }

        if (this.map == null)
        {
            String errorMessage = i18n.getMessage(
                Resources.NO_FILE_MAPPINGS_FOR_RESOURCE_MANAGER,
                manager);
            throw new FileMappingException(errorMessage);
        }
    }

    public String mapFileName(GlobusURL url)
        throws                              FileMappingException
    {
        String path = url.getPath();

        if (path == null) {
            path = "/";
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Mapping filename " + path);
        }

        try
        {
            path = URLDecoder.decode(path, URL_ENCODING);
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }

        int length = path.length();

        //make sure this isn't a Windows path before treating as a Unix path
        if (!(   (length >= 2)
              && Character.isLetter(path.charAt(0))
              && path.charAt(1) == ':'))
        {
            for (int i = 0; i < length; i++)
            {
                if (path.charAt(i) != '/')
                {
                    path = "/" + path.substring(i);
                    break;
                }
            }
        }

        FtpServerType ftpServer = map.getFtpServer();
        String protocol = ftpServer.getProtocol();
        String host = ftpServer.getHost();
        int port = ftpServer.getPort().intValue();
        FileSystemPathMappingType mapping[]= map.getMapping();
        int prefixLength = 0;
        String prefix = null;
        String replacement = null;
        StringBuffer result;

        logger.debug("Mapping length is " + Integer.toString(mapping.length));
        for (int i = 0; i < mapping.length; i++)
        {
            String tmp = mapping[i].getJobPath();

            if (logger.isDebugEnabled()) {
                logger.debug("Comparing `" + tmp
                        + "' to beginning of `" + path + "'.");
            }

            if (tmp.length() > prefixLength && path.startsWith(tmp)) {
                prefix = tmp;
                prefixLength = tmp.length();
                replacement = mapping[i].getFtpPath();
            }
        }
        if (prefix != null)
        {
            result = new StringBuffer();
            result.append(protocol);
            result.append("://");
            result.append(ftpServer.getHost());
            result.append(":");
            result.append(ftpServer.getPort());

            if (! replacement.startsWith("/"))
            {
                result.append('/');
            }

            result.append(replacement);

            if (   (replacement.length() >= 2)
                && Character.isLetter(replacement.charAt(0))
                && replacement.charAt(1) == ':')
            {
                //Windows path
                if (!
                    (   replacement.endsWith("\\")
                     || path.substring(prefixLength).startsWith("\\")))
                {
                    result.append('\\');
                }
            }
            else
            {
                //Unix path
                if (!
                    (   replacement.endsWith("/")
                     || path.substring(prefixLength).startsWith("/")))
                {
                    result.append('/');
                }
            }
            result.append(path.substring(prefixLength));

            String resultString = result.toString();

            //Change all Windows back-slashes to for-slashes
            try
            {
                resultString = resultString.replaceAll("\\\\", "%5C");
            }
            catch (Exception e)
            {
                String errorMessage = i18n.getMessage(
                    Resources.SLASH_CONVERSION_FAILED,
                    path);
                throw new FileMappingException(errorMessage, e);
            }

            if (logger.isDebugEnabled()) {
                logger.debug("Result of mapping `" + path + "' is `" +
                    resultString + "'");
            }

            return resultString;
        }
        else
        {
            String errorMessage = i18n.getMessage(
                Resources.NO_FILE_MAPPING_FOR_PATH,
                path);
            throw new FileMappingException(errorMessage);
        }
    }
}
