/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.globus.exec.generated.FileSystemMapsType;
import org.globus.exec.generated.FileSystemMapType;
import org.globus.exec.generated.FtpServerType;
import org.globus.exec.generated.FileSystemPathMappingType;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.apache.axis.utils.XMLUtils;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;


/**
 * Helper class to add, replace, remove <b>map</b> elements from an
 * xml file containing a fileSystemMaps document (typically used with
 * $GLOBUS_LOCATION/etc/gram-service/globus_gram_fS_map_config.xml).
 */
public class FileMappingConfig {
    private Document doc;
    private File path;
    static public final String DEFAULT_MAP_FILE = 
            "/etc/gram-service/globus_gram_fs_map_config.xml";
    static final private String GLOBUS_HOSTNAME_EXEC =
            "/bin/globus-hostname";
    static private java.util.Properties env;
    static private final String GRAM_NS =
                    "http://www.globus.org/namespaces/2004/10/gram/job";
    static private final javax.xml.namespace.QName MAP_QNAME =
            new javax.xml.namespace.QName(GRAM_NS, "map");

    /**
     * Create a new FileMappingConfig for manipulating a map in a file.
     * @param config Path to xml file.
     */
    public FileMappingConfig(File config) throws FileNotFoundException,
            javax.xml.parsers.ParserConfigurationException,
            org.xml.sax.SAXException,
            java.io.IOException {
        FileInputStream sourceInput = null;
        try {
            path = config;
            sourceInput = new FileInputStream(config);
            doc = XMLUtils.newDocument(sourceInput);
        } finally {
            if (sourceInput != null) {
                sourceInput.close();
            }
        }
    }

    /**
     * Get the first mapping for a particular scheduler if present.
     * @param scheduler Scheduler name to look for.
     */
    private Node getOldMap(String scheduler)
    {
        NodeList oldMapping = doc.getElementsByTagName("scheduler");
        int len = oldMapping.getLength();

        for (int i = 0; i < len; i++) {
            Node node = oldMapping.item(i);

            if (node != null) {
                String text = node.getFirstChild().getNodeValue();
                if (text != null && text.equals(scheduler)) {
                    return node.getParentNode();
                }
            }
        }
        oldMapping = doc.getElementsByTagNameNS(GRAM_NS, "scheduler");
        len = oldMapping.getLength();

        for (int i = 0; i < len; i++) {
            Node node = oldMapping.item(i);

            if (node != null) {
                String text = node.getFirstChild().getNodeValue();
                if (text != null && text.equals(scheduler)) {
                    return node.getParentNode();
                }
            }
        }
        return null;
    }

    /**
     * Remove map entries for the named scheduler.
     */
    public void removeMap(String scheduler)
            throws DOMException
    {
        Node oldMap;
        do {
            oldMap = getOldMap(scheduler);

            if (oldMap != null) {
                NodeList fileSystemMaps =
                        doc.getElementsByTagName("fileSystemMaps");
                fileSystemMaps.item(0).removeChild(oldMap);
            }
        } while (oldMap != null);
    }

    /**
     * Add a map entry for a scheduler.
     *
     * @param map FileSystemMapType containing all mappings for the
     * particular scheduler.
     */
    public void addMap(FileSystemMapType map) throws IOException,
            org.globus.wsrf.encoding.SerializationException
    {
        Element newMap = ObjectSerializer.toElement(map, MAP_QNAME);

        NodeList fileSystemMaps = doc.getElementsByTagName("fileSystemMaps");

        fileSystemMaps.item(0).appendChild(doc.importNode(newMap, true));
    }

    /**
     * Replace all existing map entries for this scheduler with the new
     * map.
     * @param map  FileSystemMapType containing all mappings for the
     * particular scheduler.
     */
    public void replaceMap(FileSystemMapType map) throws IOException,
            org.globus.wsrf.encoding.SerializationException
    {
        Element newMap = ObjectSerializer.toElement(map, MAP_QNAME);
        // Remove old mapping for this scheduler if present
        removeMap(map.getScheduler());
        NodeList fileSystemMaps = doc.getElementsByTagName("fileSystemMaps");
        fileSystemMaps.item(0).appendChild(doc.importNode(newMap, true));
    }

    /** Replace the config file on disk with the modified map.
     */
    public void writeConfig() throws IOException {
        FileOutputStream stream = null;
        try {
            stream = new FileOutputStream(path);
            XMLUtils.DocumentToStream(doc, stream);
        } finally {
            if (stream != null) {
                stream.close();
            }
        }
    }

    /**
     * Display usage message
     */
    static void usage() {
        System.out.println(
            "Usage: FileMappingConfig -scheduler SCHEDULER [options]\n"+
            "Required Parameters:\n"+
            "    -scheduler SCHEDULER                "+
            "Add mapping for named scheduler\n"+
            "Options:\n"+
            "    -file MAPPING-FILE                  "+
            "Modify MAPPING-FILE instead\n"+
            "                                        "+
            "of default\n"+
            "    -ftpServerHost HOST                 "+
            "Host name of FTP server\n"+
            "    -ftpServerPort PORT                 "+
            "Port number of FTP server [2811]\n"+
            "    -ftpProtocol PROTOCOL               "+
            "Protocol (ftp or gsiftp) [gsiftp]\n"+
            "    -mapping JOB_PATH FTP_PATH          "+
            "Add path mapping for JOB_PATH to FTP_PATH\n"+
            "    -help                               "+
            "Print this message\n"+
            "    -debug                              "+
            "Print stack trace when exceptions occur\n");
    }

    /** Determine environment values */
    static private void loadEnvironment()
            throws java.io.IOException
    {
        Runtime r = Runtime.getRuntime();

        Process envProc = r.exec("/usr/bin/env");
        env = new java.util.Properties();
        env.load(envProc.getInputStream());
    }

    /** Determine host name for default (grid)ftp server */
    static private String getHostName() throws java.io.IOException {
        Runtime r = Runtime.getRuntime();

        if (env == null) {
            loadEnvironment();
        }

        Process globusHostname = r.exec(env.getProperty("GLOBUS_LOCATION") +
                GLOBUS_HOSTNAME_EXEC);
        java.io.BufferedReader input =
                new java.io.BufferedReader(
                    new java.io.InputStreamReader(
                        globusHostname.getInputStream()));
        String hostname = input.readLine();
        input.close();
        return hostname;
    }

    static public void main(String [] args) {
        boolean debug = true;
        try {
            FileSystemMapType schedulerMap = new FileSystemMapType();

            loadEnvironment();

            File mappingFile = new File(env.getProperty("GLOBUS_LOCATION")
                    + DEFAULT_MAP_FILE);
            FtpServerType ftp = new FtpServerType();
            ftp.setProtocol("gsiftp");
            ftp.setHost(getHostName());
            ftp.setPort(new org.apache.axis.types.UnsignedShort(2811));
            schedulerMap.setFtpServer(ftp);;
            java.util.Vector mappings = new java.util.Vector();
            debug = false;

            for (int i = 0; i < args.length; i++) {
                if (args[i].equals("-scheduler") && i+1 < args.length) {
                    schedulerMap.setScheduler(args[++i]);
                } else if (args[i].equals("-file") && i+1 < args.length) {
                    mappingFile = new File(args[++i]);
                } else if (args[i].equals("-ftpServerHost") &&
                        i+1 < args.length) {
                    schedulerMap.getFtpServer().setHost(args[++i]);
                } else if (args[i].equals("-ftpServerPort") &&
                        i+1 < args.length) {
                    schedulerMap.getFtpServer().setPort(
                            new org.apache.axis.types.UnsignedShort(
                                    args[++i]));
                } else if (args[i].equals("-ftpProtocol") &&
                        i+1 < args.length) {
                    ++i;

                    if (args[i].equals("ftp") || args[i].equals("gsiftp")) {
                        schedulerMap.getFtpServer().setProtocol(args[i]);
                    } else {
                        System.err.println("Unknown protocol: " + args[i]);
                        System.exit(1);
                    }
                } else if (args[i].equals("-mapping") &&
                        i+2 < args.length) {
                    
                    mappings.add(new FileSystemPathMappingType(
                            args[++i], args[++i]));
                } else if (args[i].equals("-debug")) {
                    debug = true;;
                } else if (args[i].equals("-help")) {
                    usage();
                    System.exit(0);
                } else {
                    System.err.println("Unknown option: " + args[i] + "\n");
                    usage();
                    System.exit(1);
                }
            }

            if (schedulerMap.getScheduler() == null) {
                System.err.println(
                        "Missing required parameter -scheduler\n");
                usage();
                System.exit(1);
            }

            if (mappings.isEmpty()) {
                    mappings.add(new FileSystemPathMappingType("/", "/"));
            }
            schedulerMap.setMapping(
                (FileSystemPathMappingType[]) mappings.toArray(
                    new FileSystemPathMappingType[0]));

            FileMappingConfig c = new FileMappingConfig(mappingFile);
            c.replaceMap(schedulerMap);
            c.writeConfig();
        } catch(Exception e) {
            System.err.println("Unable to update map file" + e.toString());
            if (debug) {
                e.printStackTrace();
            }
        }
    }
}
