/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Map;
import java.util.HashMap;
import javax.security.auth.Subject;
import javax.xml.rpc.holders.IntHolder;
import javax.xml.namespace.QName;
import javax.naming.Context;
import javax.naming.InitialContext;
import org.apache.axis.AxisEngine;
import org.apache.axis.providers.java.JavaProvider;
import org.apache.axis.MessageContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceThread;
import org.globus.wsrf.container.ServiceManager;
import org.globus.axis.utils.MessageContextHelper;
import org.globus.wsrf.utils.ContextUtils;
import org.globus.wsrf.impl.lifetime.DestroyProvider;
import org.globus.wsrf.impl.notification.SubscribeHelper;
import org.globus.wsrf.impl.properties.GetResourcePropertyProvider;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.globus.gsi.jaas.UserNamePrincipal;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.DestroyResponse;
import org.oasis.wsn.Subscribe;
import org.apache.axis.message.addressing.EndpointReferenceType;

public class LocalInvocationHelper
{  
    private static Log logger = 
        LogFactory.getLog(LocalInvocationHelper.class.getName());
    
    private static LocalInvocationHelper INSTANCE = null;
    private AxisEngine axisEngine;
    private Map serviceCache;
    private Map methodCache;
    private DestroyProvider destroyProvider;
    private GetResourcePropertyProvider getRPProvider;
    private String configDir;
    private String homeDir;

    /**
     * Private constructor. Get the object of 
     * LocalInvocationHelper via getInstance()
     */
    private LocalInvocationHelper ()
    {
        axisEngine = ContainerConfig.getEngine();
        this.configDir = ServiceThread.getConfigRootPath(axisEngine);
        try {
            this.homeDir = ServiceThread.getWebRootPath(axisEngine);
        } catch (Exception e) {
            logger.warn("",e);
        }
        
        serviceCache = new HashMap();
        methodCache = new HashMap();

        // Get an instance of the DestroyProvider and the GetResourcePropertyProvider
        destroyProvider = new DestroyProvider();
        getRPProvider = new GetResourcePropertyProvider();
    }
    
    /**
     * Get the instance of the LocalInvocationHelper 
     * @return object of LocalInvocationHelper
     */
    public static synchronized LocalInvocationHelper getInstance()
    {
        if (INSTANCE == null) {
            INSTANCE = new LocalInvocationHelper();
        }
        return INSTANCE;
    }

    /**
     * Call a method of a service locally without doing a WS call
     *
     * @param serviceName Name of the service that will be called
     * @param wsddFileName Name of the deployment descriptor file,
              relative to $GLOBUS_LOCATION
     * @param methodName Name of the method that will be called
     * @param userSubject Subject of the user who does the call
     * @param args Arguments passed to the method that will be called
     * @param resourceKey key of the resource, the method will work on
     * @return Object Whatever the call returns
     * @throws Exception
     */
    public Object callService (
          String                           serviceName,
          String                           methodName,
          String                           userSubject,
          Object[]                         args,
          ResourceKey                      resourceKey) 
          throws                           Exception
    {
        MessageContextBundle msgCtxBundle = null;
      
        // do the local call
        Object returnValue = null;
        try {
            msgCtxBundle = prepareMessageContext(
                             serviceName,
                             resourceKey,
                             userSubject);
                             
            // get the method that will be called
            Method method = getMethod(
                              msgCtxBundle.getMessageContext(),
                              serviceName,
                              methodName);
                              
            logger.debug("calling " + serviceName + 
                         " (method " + methodName + ")");
            returnValue = method.invoke(
                             msgCtxBundle.getServiceObject(),
                             args);
        } catch (InvocationTargetException e) {
            throw (Exception) e.getCause();
        } finally {
            if (msgCtxBundle != null) {
               // de-associate the MessageContext from the current thread
               msgCtxBundle.getMessageContextHelper().unset();
            }
        }
        
        return returnValue;
    }

    /**
     *  Call the DestroyProvider to remove a resource.
     *
     * @param serviceName name of the service
     * @param resourceKey key of the resource that will be destroyed
     * @param userSubject subject of the user who does the invocation
     * @return DestroyResponse 
     * @throws Exception
     */
    public DestroyResponse destroyResource (
          String                                  serviceName,
          ResourceKey                             resourceKey,
          String                                  userSubject) 
          throws                                  Exception
    {       
        DestroyResponse response = null;
        MessageContextBundle msgCtxBundle = null;
        
        // do the local call
        try {
            msgCtxBundle = prepareMessageContext(
                             serviceName,
                             resourceKey,
                             userSubject);
            response = destroyProvider.destroy(new Destroy());
        } catch (Exception e) {
            throw(e);
        } finally {
            if (msgCtxBundle != null) {
               // de-associate the MessageContext from the current thread
               msgCtxBundle.getMessageContextHelper().unset();
            }
        }
            
        return response;
    }

    /**
     *  Call the GetResourcePropertyProvider to get a RP
     *
     * @param serviceName name of the service
     * @param resourceKey key of the resource that will be destroyed
     * @param qname QName of the RP
     * @param userSubject subject of the user who does the invocation
     * @return DestroyResponse 
     * @throws Exception
     */
    public GetResourcePropertyResponse getResourceProperty (
          String                                  serviceName,
          ResourceKey                             resourceKey,
          QName                                   qname,
          String                                  userSubject) 
          throws                                  Exception
    {       
        GetResourcePropertyResponse response = null;
        MessageContextBundle msgCtxBundle = null;

        // do the local call
        try {
            msgCtxBundle = prepareMessageContext(
                             serviceName,
                             resourceKey,
                             userSubject);
            logger.debug("getting resource property");
            response = getRPProvider.getResourceProperty(qname);
        } catch (Exception e) {
            throw(e);
        } finally {
            if (msgCtxBundle != null) {
               // de-associate the MessageContext from the current thread
               msgCtxBundle.getMessageContextHelper().unset();
            }
        }
        
        return response;
    }
    
    /**
     * Subscribe for locally sent notifications. 
     *
     * @param serviceName Name of the service
     * @param resourceKey resource key of the notification producer resource
     * @param userSubject subject of the user who does the invocation
     * @param subscribeRequest subscribe request input
     * @param callback callback whose deliver method will be called
     * @return EndpointReferenceType epr of the subscription resource
     * @throws Exception
     */
    public EndpointReferenceType subscribe(
           String                             serviceName,
           ResourceKey                        resourceKey,
           String                             userSubject,
           Subscribe                          subscribeRequest,
           NotifyCallback                     callback)
           throws                             Exception
    {
        ResourceKey  producerKey = resourceKey;
        EndpointReferenceType  response = null;
        MessageContextBundle msgCtxBundle = null;

        try {
           msgCtxBundle = prepareMessageContext(
                             serviceName,
                             resourceKey,
                             userSubject);
           Context initialContext = new InitialContext();
           String producerHomeLocation = 
                org.globus.wsrf.Constants.JNDI_SERVICES_BASE_NAME + 
                serviceName + org.globus.wsrf.Constants.HOME_NAME;
           ResourceHome producerHome = 
                (ResourceHome) initialContext.lookup(producerHomeLocation);
           Resource producerResource = producerHome.find(resourceKey);
           SubscribeHelper subscribeHelper 
              = new SubscribeHelper(
                       producerResource,
                       producerKey,
                       producerHomeLocation,
                       subscribeRequest,
                       callback);
            response = subscribeHelper.subscribe();
        } catch (Exception e) {
            throw(e);
        } finally {
            if (msgCtxBundle != null) {
               // de-associate the MessageContext from the current thread
               msgCtxBundle.getMessageContextHelper().unset();
            }
        }

        return response;
    }
    

     /**
      * Create a message context, associate the message context to the current
      * thread and load the service object.
      *
      * @param serviceName name of the service that will be called
      * @param resourceKey resource key or null if no one is needed
      * @userSubject subject of the user who does the call
      * @return MessagContextHelper to de-associate the message from the thread
      * @throws Exception
      */
     public MessageContextBundle prepareMessageContext (
          String            serviceName,
          ResourceKey       resourceKey,
          String            userSubject)
          throws            Exception
     {
         MessageContext msgCtx = null;
         MessageContextHelper msgCtxHelper = null;
         Object serviceObject = null;
         
         // create MessageContext
         msgCtx = createMessageContext(
                      serviceName,
                      resourceKey,
                      userSubject);

         // associate the MessageContext to the current thread
         msgCtxHelper = new MessageContextHelper(msgCtx);
         msgCtxHelper.set();

         // get the service object corresponding to the service name.
         // even if the service object is not needed directly, the service gets 
         // loaded if it had not been loaded already. This ensures that potential
         // initialization gets done in the service constructor which may be
         // needed even if an operation provider will be called
         ServiceManager.initializeService(msgCtx);
         serviceObject = getServiceObject(
                             serviceName,
                             msgCtx);

         return new MessageContextBundle(
                         serviceObject,
                         msgCtx,
                         msgCtxHelper);
     }


     
     /**
      * Helper method to create a MessageContext and populate the necessary
      * properties for use with local invocations.
      * 
      * @param serviceName Name of the service that will be called
      * @param resourceKey Resource key that must be added to the MessageContext
      * @param userSubject Subject that must be added to the MessageContext
      * @return msgCtx Generated MessageContext
      * @throws Exception
      */
     public MessageContext createMessageContext (
               String                 serviceName,
               ResourceKey            resourceKey,
               String                 userSubject)
               throws                 Exception
     {
         // message context bound to the calling thread
         MessageContext msgCtx = null;
         
         // create the MessageContext for the local call
         Subject subject = new Subject();
         subject.getPrincipals().add(new UserNamePrincipal(userSubject));
         msgCtx = new MessageContext(axisEngine);
         msgCtx.setTargetService(serviceName);
         msgCtx.setProperty(
              org.globus.wsrf.impl.security.authentication.Constants.PEER_SUBJECT,
              subject);
         msgCtx.setProperty(
              org.globus.wsrf.Constants.LOCAL_INVOCATION,
              new Boolean(true));
         if (resourceKey != null) {
             msgCtx.setProperty(
                 org.globus.wsrf.Constants.LI_RESOURCE_KEY,
                 resourceKey);
         }    
         msgCtx.setProperty(org.apache.axis.Constants.MC_HOME_DIR, this.homeDir);
         msgCtx.setProperty(org.apache.axis.Constants.MC_CONFIGPATH, this.configDir);
               
         return msgCtx;
     }


    /**
     * Get the service object. Class name and Provider are looked up in
     * the deployment descriptor which is passed as argument.
     * Once a service object has been created it's cached for further use.
     * NOTE: This method must be called with a valid message context
     * associated to the current thread. 
     *
     * @param serviceName Name of the service to be called
     * @param msgCtx MessageContext of the current thread
     * @return Object Object of the requested service
     * @throws Exception 
     */
    public synchronized Object getServiceObject (
           String           serviceName,
           MessageContext   msgCtx)
           throws           Exception
    {

        Object serviceObject = null;
         
        // check if ServiceObject is cached
        if (serviceCache.containsKey(serviceName)) {
            logger.debug("getting cached service object");
            serviceObject = serviceCache.get(serviceName);
        } else {
            String serviceClassName = 
                (String) ContextUtils.getServiceProperty(
                                  msgCtx,
                                  serviceName,
                                  "className");
            String providerClassName =
                (String) ContextUtils.getServiceProperty(
                                  msgCtx,
                                  serviceName,
                                  "handlerClass");
        
            logger.debug("className="+serviceClassName+" "+
                         "handlerClass="+providerClassName);
            
            // create a Provider object
            Class providerClass = Class.forName(providerClassName);
            Constructor co = providerClass.getConstructor(null);
            JavaProvider provider = (JavaProvider) co.newInstance(null);
            
            // get the service object
            logger.debug("getting the service object");
            serviceObject =
                provider.getServiceObject(
                        msgCtx,
                        msgCtx.getService(),
                        serviceClassName,
                        new IntHolder());
                        
            // cache service object
            logger.debug("caching service object for service "+serviceName);
            serviceCache.put(serviceName,serviceObject);
        }        
        return serviceObject;
    }
    
    /**
     * Get a method of the service. Check first in the methodCache, whether
     * it is cached there before creating it via reflection
     *
     * @param msgCtx MessageContext
     * @param serviceName Name of the service that will be called
     * @param methodName Name of the method that will be called
     * @return Method Object of the requested method
     * @throws Exception
     */
     public synchronized Method getMethod (
             MessageContext              msgCtx,
             String                      serviceName,
             String                      methodName)
             throws Exception
     {
         Method method = null;
         String name = serviceName+"."+methodName;
         if (methodCache.containsKey(name)) {
           method = (Method) methodCache.get(name);
         } else {  
           method = msgCtx.getService()
                           .getServiceDescription()
                            .getOperationByName(methodName)
                             .getMethod();
           logger.debug("caching method "+name);
           methodCache.put(name,method);
         }
         return method;
     }


     public class MessageContextBundle
     {
         private MessageContext msgCtx = null;
         private MessageContextHelper msgCtxHelper = null;
         private Object serviceObject = null;

         public MessageContextBundle(
             Object                 serviceObject,
             MessageContext         msgCtx,
             MessageContextHelper   msgCtxHelper)
         {
             this.serviceObject = serviceObject;
             this.msgCtx = msgCtx;
             this.msgCtxHelper = msgCtxHelper;
         }

         public Object getServiceObject()
         {
             return this.serviceObject;
         }

         public MessageContext getMessageContext()
         {
             return this.msgCtx;
         }

         public MessageContextHelper getMessageContextHelper()
         {
             return this.msgCtxHelper;
         }
     }

}

