/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.factory;

import java.io.File;
import java.io.FileInputStream;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import javax.naming.InitialContext;
import javax.xml.namespace.QName;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

import org.globus.exec.utils.Resources;
import org.globus.exec.monitoring.JobStateMonitor;
import org.globus.exec.monitoring.JobStateChangeListener;
import org.globus.exec.monitoring.JobStateRecoveryListener;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.service.exec.ManagedExecutableJobHome;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertyMetaData;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.impl.ReflectionResource;
import org.globus.wsrf.impl.security.descriptor.GSISecureConvAuthMethod;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.security.SecureResource;
import org.globus.mds.usefulrp.ServiceMetaDataResourceProperty;
import org.globus.mds.usefulrp.rpprovider.GLUEResourcePropertyConfig;
import org.globus.mds.usefulrp.rpprovider.ResourcePropertyProviderCollection;


public class ManagedJobFactoryResource extends      ReflectionResource
                                       implements   SecureResource,
                                                    JobStateRecoveryListener
{
    private static Log logger
        = LogFactory.getLog(ManagedJobFactoryResource.class);

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static Timer SUB_DEF_REFRESH_TIMER = new Timer(true);

    protected Properties substitutionDefinitions = null;

    private JobStateMonitor monitor;
    private File timestampPath;
        
    /*
    private static final String SEG_EXECUTABLE_NAME =
        "globus-scheduler-event-generator";
        */

    // the GLUE resource property data generation is handled by this object
    private ResourcePropertyProviderCollection glueRP;

    private boolean enableDefaultSoftwareEnvironment;

    public void initialize(
        Object                              resourceBean,
        QName                               resourceElementQName,
        Object                              key,
        String                              subDefFile,
        String                              subDefRefreshPeriod,
        String                              enableDefaultSoftEnv)
        throws                              ResourceException
    {
        super.initialize(resourceBean, resourceElementQName, key);

        //Let's derive from ReflectionResource so we can add
        //specialized behavior here:

        String globusLocation = getStringProperty(
            ManagedJobFactoryConstants.RP_GLOBUS_LOCATION);

        if (globusLocation == null) {
            String errorMessage =
                i18n.getMessage(Resources.RP_NOT_FOUND,
                                ManagedJobFactoryConstants.RP_GLOBUS_LOCATION);
            throw new ResourceException(errorMessage);
        }

        /* Parameters to JobStateMonitor constructor: seg path, seg user
         * name, scheduler module path, state change listener, recovery
         * listener
         */

        /* SEG name is
         * $GLOBUS_LOCATION/libexec/globus-scheduler-event-generator
         */
        /*
        File segPath = new File(globusLocation +
                File.separator + "libexec" +
                File.separator + SEG_EXECUTABLE_NAME);
                */

        /* TODO: SEG user name. This is supposed to be the name which the
         * separate SEG process will run as. This is currently unimplemented.
         */
        String segUserName = "dummy";

        /* SEG scheduler name is the name of the scheduler we are using
         */
        String schedulerName = getStringProperty(
            ManagedJobFactoryConstants.RP_LOCAL_RESOURCE_MANAGER);

        if (schedulerName == null) {
            String errorMessage =
                i18n.getMessage(
                    Resources.RP_NOT_FOUND,
                    ManagedJobFactoryConstants.RP_LOCAL_RESOURCE_MANAGER);
            throw new ResourceException(errorMessage);

        }
        /* Convert to lower case to reach naming convention of SEG modules */
        schedulerName = schedulerName.toLowerCase();

        /* State Change Listener is the service's ResourceHome */
        ManagedExecutableJobHome home = ManagedExecutableJobHome.getHome();

        if (! (home instanceof JobStateChangeListener)) {
            String errorMessage = i18n.getMessage(Resources.INTERNAL_ERROR);
            throw new ResourceException(errorMessage);
        }

        JobStateChangeListener jobStateChangeListener =
                (JobStateChangeListener) home;

        /* Recovery listener is this */
        JobStateRecoveryListener recoveryListener = this;

        timestampPath = new File(globusLocation + File.separator +
                "var" + File.separator +
                "globus-jsm-" + schedulerName + ".stamp");

        if (!schedulerName.equals("multi"))
        {
            this.monitor = JobStateMonitor.getInstance(
                new File(globusLocation),
                segUserName,
                schedulerName,
                jobStateChangeListener,
                recoveryListener,
                true);

            this.monitor.start(getRestartTimestamp());
        }

        //schedule periodic substitution definitions refresh task
        long refreshPeriod;
        try
        {
            refreshPeriod = Long.parseLong(subDefRefreshPeriod) * 60000;
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
        SUB_DEF_REFRESH_TIMER.schedule(
            new SubstitutionDefinitionsRefreshTask(this, subDefFile),
            refreshPeriod,
            refreshPeriod);

        //default software environment flag
        this.enableDefaultSoftwareEnvironment
            = new Boolean(enableDefaultSoftEnv).booleanValue();

        //recover jobs
        class RecoveryThread extends Thread
        {
            ManagedExecutableJobHome home;

            public RecoveryThread(ManagedExecutableJobHome home) {
                super();
                this.home = home;
            }
            public void run() {
                ManagedJobFactoryResource.logger.debug(
                    "recovery thread active");
                try {
                    synchronized (home)
                    {
                        while (!home.isInitialized())
                        {
                            try { home.wait(); } catch (Exception e) { }
                        }
                        home.recover();
                    }
                } catch (org.globus.wsrf.ResourceException re) {
                    logger.warn("Recovery exception", re);
                }
                ManagedJobFactoryResource.logger.debug(
                    "recovery thread finished");
            }
        };

        Thread recoveryThread = new RecoveryThread(home);

        recoveryThread.start();
        
        this.createNewGLUEResourceProperty();
    }

    private String getStringProperty(QName propertyName) {
        ResourcePropertySet props = getResourcePropertySet();
        return (String) props.get(propertyName).get(0);
    }

    /**
     * Implementation of the JobStateRecoveryListener interface.
     *
     * This method is called periodically to allow the factory resource to
     * store the JobStateMonitor's recovery timestamp in a persistent
     * (non-resource property) store. This timestamp is reread when the
     * service container is restarted to allow the JobStateMonitor to resume
     * generating events from a point where it won't lose events.
     *
     * @param monitor
     *     The monitor which this update pertains to. This should match
     *     the monitor referred to by this resource.
     * @param timeStamp
     *     Timestamp from which it is safe to recover from.
     */
    public void updateJobMonitorRecoveryTimeStamp(
        JobStateMonitor                     monitor,
        java.util.Date                      timeStamp)
    {
        try {
            if (monitor != this.monitor) {
                /* something is broken */
                return;
            }

            java.io.ObjectOutputStream oos = new java.io.ObjectOutputStream(
                    new java.io.FileOutputStream(timestampPath));

            oos.writeObject(timeStamp);
            oos.close();
        } catch (java.io.IOException ioe) {
            /* Not much we can do here. */
            logger.warn(ioe);
        }
    }

    /**
     * Reload the restart JSM timestamp from a well-known file.
     *
     * @returns A reference to the last timestamp stored in the recovery file,
     *     or null if no such timestamp can be read.
     */
    private java.util.Date getRestartTimestamp() {
        try {
            java.io.ObjectInputStream ois = new java.io.ObjectInputStream(
                    new java.io.FileInputStream(timestampPath));
            if (ois == null) {
                return null;
            }
            Object o = ois.readObject();

            ois.close();

            if (o == null || ! (o instanceof java.util.Date)) {
                return null;
            }
            else {
                return (java.util.Date) o;
            }
        } catch (java.io.IOException ioe) {
            logger.debug(ioe);

            return null;
        } catch (ClassNotFoundException cnfe) {
            logger.warn(cnfe);

            return null;
        }
    }

    /**
     * Return the JobStateMonitor associated with this scheduler resource.
     */
    public JobStateMonitor getJobStateMonitor() {
        return this.monitor;
    }

    //********** Security Descriptor Fields and Methods **********//

    private ResourceSecurityDescriptor securityDescriptor = null;

    public ResourceSecurityDescriptor getSecurityDescriptor() {
        return this.securityDescriptor;
    }


    /** Returns custom constructed RPs for some RPs, passes up to
        superclass for all the rest. */
    protected ResourceProperty createNewResourceProperty(
                                       ResourcePropertyMetaData metaData,
                                       Object resourceBean)
    throws Exception
    {
        QName rpQName = metaData.getName();
        logger.debug("Creating resource property for: "+rpQName);

        if(rpQName.equals(
            ServiceMetaDataResourceProperty.SERVICE_METADATA ))
        {
            return createNewMDSMetaDataResourceProperty(rpQName);
        } else if(rpQName.equals(
            ManagedJobFactoryConstants.RP_DELEGATION_FACTORY_ENDPOINT))
        {
            return DelegationFactoryEndpointResourceProperty.getInstance();
        } else if(rpQName.equals(
            ManagedJobFactoryConstants.RP_STAGING_DELEGATION_FACTORY_ENDPOINT))
        {
            return StagingDelegationFactoryEndpointResourceProperty
                   .getInstance();
        } else
        {
            return super.createNewResourceProperty(metaData, resourceBean);
        }
    }

    protected ResourceProperty createNewMDSMetaDataResourceProperty(
        QName                               rpQName)
    {
        ResourceProperty metadata = null;
        try {
            metadata=new ServiceMetaDataResourceProperty();
        } catch(Exception e) {
            logger.debug("Unable to add ServiceMetaData RP");
        }
        return metadata;
    }

    protected void createNewGLUEResourceProperty()
    {
        try 
        {
            String globusLocation = getStringProperty(
                ManagedJobFactoryConstants.RP_GLOBUS_LOCATION);

            String schedulerName = getStringProperty(
                ManagedJobFactoryConstants.RP_LOCAL_RESOURCE_MANAGER);

            if (schedulerName == null) {
                logger.warn(
                    "No scheduler name found while attempting to create GLUE RP - resource property will not be available");
                return;
            }

            // no meaningful GLUE representation for multi
            if (schedulerName.equalsIgnoreCase("Multi")) {
                return;
            }
            
            String glueConfigFile = 
                globusLocation + File.separator + "etc/gram-service-" + 
                schedulerName + File.separator + "gluerp-config.xml";

            File configFile = new File(glueConfigFile);

            if (!(new File(glueConfigFile).exists())) {
                GLUEResourcePropertyConfig.generateDefaultConfig(
                    glueConfigFile, schedulerName, null);
            }

            this.glueRP = 
                new ResourcePropertyProviderCollection(
                    glueConfigFile,
                    this.getResourcePropertySet(),
                    true);            
        }
        catch (Exception e) 
        {
            logger.warn("Exception while creating GLUE Resource Property: "+ e);
        }

    }
    
    public EndpointReferenceType getStagingDelegationFactoryEndpoint()
    {
        ResourcePropertySet props = getResourcePropertySet();
        return (EndpointReferenceType) props.get(
            ManagedJobFactoryConstants.RP_STAGING_DELEGATION_FACTORY_ENDPOINT)
            .get(0);
    }

    public boolean isDefaultSoftwareEnvironmentEnabled()
    {
        return this.enableDefaultSoftwareEnvironment;
    }

    public Set substitutionDefinitionSet()
    {
        return this.substitutionDefinitions.entrySet();
    }

    class SubstitutionDefinitionsRefreshTask extends TimerTask
    {
        ManagedJobFactoryResource factoryResource = null;
        String subDefFile = null;

        public SubstitutionDefinitionsRefreshTask(
            ManagedJobFactoryResource       factoryResource,
            String                          subDefFile)
        {
            this.factoryResource = factoryResource;
            this.subDefFile = subDefFile;

            //make sure the variables are set to avoid a race condition
            refresh();
        }

        private void refresh()
        {
            try
            {
                factoryResource.substitutionDefinitions = new Properties();
                factoryResource.substitutionDefinitions.load(
                    new FileInputStream(this.subDefFile));
            }
            catch (Exception e)
            {
                throw new RuntimeException(e);
            }

            if (logger.isDebugEnabled())
            {
                logger.debug("Refreshed Substitution Definitions:\n"
                    + factoryResource.substitutionDefinitions.toString());
            }
        }

        public void run()
        {
            refresh();
        }
    }
}
