/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.job;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.security.GeneralSecurityException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;

import javax.xml.namespace.QName;

import org.apache.axis.message.addressing.EndpointReferenceType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

import org.globus.delegation.DelegationException;

import org.globus.exec.utils.Resources;
import org.globus.exec.generated.FaultResourcePropertyType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.FilePairType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedJobResourceDataType;
import org.globus.exec.generated.MultiJobDescriptionType;
import org.globus.exec.generated.ServiceLevelAgreementType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.monitoring.JobStateMonitor;
import org.globus.exec.service.utils.DelegatedCredential;
import org.globus.exec.utils.ManagedJobConstants;

import org.globus.rendezvous.service.RendezvousResourceImpl;

import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import org.oasis.wsrf.faults.BaseFaultType;

public abstract class ManagedJobResourceImpl
    extends                                 RendezvousResourceImpl
    implements                              ManagedJobResource
{
    private Log logger = null;
    private Log localLogger = null;

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private ManagedJobResourceDataType resourceData;

    //delegated credentials
    private DelegatedCredential jobCredential = null;
    private DelegatedCredential stagingCredential = null;
    private static String JOB_CREDENTIAL_TYPE = "job";
    private static String STAGING_CREDENTIAL_TYPE = "staging";

    public void initializeWholeState(
        Object                                  wholeResourceState,
        QName                                   resourcePropertiesQName,
        Object                                  resourceKey)
        throws                                  ResourceException
    {
        this.logger = LogFactory.getLog(
            ManagedJobResourceImpl.class.getName()
            + "." + ((ResourceKey)resourceKey).getValue().toString());
    }

    /**
     * @param name QName
     * @return Object the value(s) of the resource property. Is a List
     *         if the cardinality of the resource property is > 1
     */
    public Object getResourceProperty(
            QName                           name) {
        ResourcePropertySet resourceProperties = getResourcePropertySet();
        ResourceProperty resourceProperty = resourceProperties.get(name);

        int size = resourceProperty.size();

        //this.logger.debug("cardinality of resource property: " +
        //             resourceProperty.getMaxOccurs());
        if (resourceProperty.getMetaData().getMaxOccurs() > 1) {

            ArrayList values = new ArrayList(size);
            for (int i = 0; i < size; i++) {
                values.add(resourceProperty.get(i));
            }
            return values;
            //we can't do values.toArray(array) and cast the result, because
            //we don't know the Java type. Let the client do it.
        }

        if (size == 0) {
            return null;
        }

        // size == 1
        return resourceProperty.get(0);

    }

    public void setResourceProperty(
            QName                           name,
            boolean                         booleanValue) {
        setResourceProperty(name, new Boolean(booleanValue));
    }

    public void setResourceProperty(
        QName                               name,
        Object                              value)
    {
        ResourceProperty resourceProperty = getResourcePropertySet().get(name);
        if (resourceProperty.size() < 1) {
            resourceProperty.add(value);
        }
        resourceProperty.set(0, value);
    }

    public static void setResourceDatum(
        ManagedJobResource                  resource,
        String                              name,
        Object                              value)
    {
        Log localLogger = LogFactory.getLog(
            ManagedJobResourceImpl.class.getName()
            + "." + ((ResourceKey)resource.getID()).getValue().toString());

        if (localLogger.isDebugEnabled())
        {
            localLogger.debug(
                "Setting resource datum " + name + " to value " + value);
        }

        ManagedJobResourceDataType resourceData = resource.getResourceData();
        localLogger.trace("Obtaining lock on resourceData");
        synchronized (resourceData)
        {
            localLogger.trace("Obtained lock on resourceData");

            BeanInfo beanInfo = null;
            try {
                beanInfo
                    = Introspector.getBeanInfo(resourceData.getClass());
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_DATA_INSTROSPECTION_ERROR,
                    resourceData.getClass().getName());
                throw new RuntimeException(errorMessage, e);
            }
            PropertyDescriptor[] propertyDescriptors
                = beanInfo.getPropertyDescriptors();

            boolean foundProperty = false;
            PropertyDescriptor propertyDescriptor = null;
            for (int index=0;
                 (index<propertyDescriptors.length) && (!foundProperty);
                 index++)
            {
                propertyDescriptor = propertyDescriptors[index];
                if (propertyDescriptor.getName().equals(name)) {
                    foundProperty = true;
                }
            }

            if (!foundProperty) {
                String errorMessage =
                    i18n.getMessage(Resources.RESOURCE_DATUM_NOT_FOUND, name);
                throw new RuntimeException(errorMessage);
            }

            Method writeMethod = propertyDescriptor.getWriteMethod();
            try {
                writeMethod.invoke(resourceData, new Object[] { value });
            } catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.RESOURCE_DATUM_SET_ERROR, name);
                throw new RuntimeException(errorMessage);
            }
            localLogger.trace("Releasing lock on resourceData");
        }
        localLogger.trace("Released lock on resourceData");
    }

    public void setResourceDatum(
            String                          name,
            boolean                         booleanValue) {
        setResourceDatum(name, new Boolean(booleanValue));
    }

    public void setResourceDatum(
        String                              name,
        Object                              value)
    {
        setResourceDatum(this, name, value);
    }

    public Object getResourceDatum(
        String                              name)
    {
        if (logger.isDebugEnabled()) {
            this.logger.debug("getting resource datum " + name);
        }

        Object value = null;

        ManagedJobResourceDataType resourceData = getResourceData();
        this.logger.trace("Obtaining lock on resourceData");
        synchronized (resourceData)
        {
            this.logger.trace("Obtained lock on resourceData");
            BeanInfo beanInfo = null;
            try {
                beanInfo
                    = Introspector.getBeanInfo(resourceData.getClass());
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.RESOURCE_DATA_INSTROSPECTION_ERROR,
                    resourceData.getClass().getName());
                throw new RuntimeException(errorMessage, e);
            }
            PropertyDescriptor[] propertyDescriptors
                = beanInfo.getPropertyDescriptors();

            boolean foundProperty = false;
            PropertyDescriptor propertyDescriptor = null;
            for (int index=0;
                 (index<propertyDescriptors.length) && (!foundProperty);
                 index++)
            {
                propertyDescriptor = propertyDescriptors[index];
                if (propertyDescriptor.getName().equals(name)) {
                    foundProperty = true;
                }
            }

            if (!foundProperty) {
                String errorMessage =
                    i18n.getMessage(Resources.RESOURCE_DATUM_NOT_FOUND, name);
                throw new RuntimeException(errorMessage);
            }

            Method readMethod = propertyDescriptor.getReadMethod();
            try {
                value = readMethod.invoke(resourceData, null);
            } catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.RESOURCE_DATUM_GET_ERROR, name);
                throw new RuntimeException(errorMessage);
            }
            this.logger.trace("Releasing lock on resourceData");
        }
        this.logger.trace("Released lock on resourceData");

        return value;
    }

    public JobDescriptionType getJobDescription() {
        ServiceLevelAgreementType sla
            = (ServiceLevelAgreementType) this.getResourcePropertySet()
                .get(ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT).get(0);
        JobDescriptionType jobDescription = sla.getJob();
        if (jobDescription == null) {
            jobDescription = sla.getMultiJob();
        }

        return jobDescription;
    }

    protected void setJobDescription(JobDescriptionType jobDescription) {
        ServiceLevelAgreementType sla
            = (ServiceLevelAgreementType) this.getResourcePropertySet()
                .get(ManagedJobConstants.RP_SERVICE_LEVEL_AGREEMENT).get(0);

        if (jobDescription instanceof MultiJobDescriptionType) {
            sla.setMultiJob((MultiJobDescriptionType)jobDescription);
        } else {
            sla.setJob(jobDescription);
        }
    }

    public FaultType getFault()
    {
        ResourceProperty faultRP
            = this.getResourcePropertySet().get(ManagedJobConstants.RP_FAULT);
        FaultResourcePropertyType faultWrapper
            = (FaultResourcePropertyType) faultRP.get(0);

        if (faultWrapper == null)
        {
            return null;
        }

        FaultType fault = null;
        if ((fault=faultWrapper.getFault()) != null)
        { } else
        if ((fault=faultWrapper.getCredentialSerializationFault()) != null)
        { } else
        if ((fault=faultWrapper.getInsufficientCredentialsFault()) != null)
        { } else
        if ((fault=faultWrapper.getInternalFault()) != null)
        { } else
        if ((fault=faultWrapper.getInvalidCredentialsFault()) != null)
        { } else
        if ((fault=faultWrapper.getServiceLevelAgreementFault()) != null)
        { } else
        if ((fault=faultWrapper.getExecutionFailedFault()) != null)
        { } else
        if ((fault=faultWrapper.getFilePermissionsFault()) != null)
        { } else
        if ((fault=faultWrapper.getInvalidPathFault()) != null)
        { } else
        if ((fault=faultWrapper.getStagingFault()) != null)
        { } else
        if ((fault=faultWrapper.getUnsupportedFeatureFault()) != null)
        { }

        return fault;
    }

    public void setFault(FaultType fault) {
        ResourceProperty faultRP
            = this.getResourcePropertySet().get(ManagedJobConstants.RP_FAULT);

        /** get appropriate fault wrapper element write method **/
        Class faultClass = fault.getClass();
        String faultClassName = faultClass.getName();
        //remove package name
        String faultElementName = faultClassName.substring(
            faultClassName.lastIndexOf('.')+1);
        this.logger.debug("fault element name: " + faultElementName);
        //remove "Type" from end of class name
        faultElementName
            = faultElementName.substring(0, faultElementName.length()-4);
        this.logger.debug("fault element name: " + faultElementName);
        //decapitalize
        faultElementName = Introspector.decapitalize(faultElementName);
        this.logger.debug("fault element name: " + faultElementName);
        Method writeMethod = null;
        try {
            PropertyDescriptor propertyDescriptor = new PropertyDescriptor(
                faultElementName, FaultResourcePropertyType.class);
            writeMethod = propertyDescriptor.getWriteMethod();
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.FAULT_WRITE_METHOD_NOT_FOUND, faultClassName);
            throw new RuntimeException(errorMessage, e);
        }


        //populate the correct element in the fault wrapper
        FaultResourcePropertyType faultWrapper
             = new FaultResourcePropertyType();
        try {
            writeMethod.invoke(faultWrapper, new Object[] { fault });
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.FAULT_WRITE_METHOD_ERROR, faultClassName);
            throw new RuntimeException(errorMessage, e);
        }

        //set the fault RP
        faultRP.set(0, faultWrapper);
    }

    //credential fetching methods
    public DelegatedCredential getJobCredential(
        File                                userProxyDirectory)
        throws                              ResourceException,
                                            DelegationException,
                                            GeneralSecurityException,
                                            IOException
    {
        if (logger.isTraceEnabled())
        {
            this.logger.trace("Entering getJobCredential()");
        }

        if (this.jobCredential == null) {
            JobDescriptionType jobDescription = getJobDescription();

            String localUserId
                = (String) this.getResourcePropertySet().get(
                    ManagedJobConstants.RP_LOCAL_USER_ID).get(0);

            EndpointReferenceType credentialEndpoint
                = jobDescription.getJobCredentialEndpoint();
            if (credentialEndpoint == null)
            {
                if (logger.isTraceEnabled())
                {
                    this.logger.trace("Leaving getJobCredential()");
                }

                return null;
            }
            if (logger.isDebugEnabled()) {
                this.logger.debug("job credential endpoint:\n"
                            + credentialEndpoint);
            }

            String userSubject
                = (String) getResourceProperty(
                    ManagedJobConstants.RP_USER_SUBJECT);

            this.jobCredential
                = DelegatedCredential.getDelegatedCredential(
                    credentialEndpoint,
                    localUserId,
                    userSubject,
                    userProxyDirectory);

            if (logger.isDebugEnabled()) {
                this.logger.debug("Delegated Job Credential Object: "
                            +this.jobCredential.toString());
                try {
                    this.logger.debug(
                        "Using proxy with subject "
                        +this.jobCredential.getCredential().getIdentity());
                } catch (Exception e) {
                    String errorMessage = i18n.getMessage(
                        Resources.DELEGATED_PROXY_SUBJECT_ERROR);
                    this.logger.error(errorMessage);
                }
            }
        }

        if (logger.isTraceEnabled())
        {
            this.logger.trace("Leaving getJobCredential()");
        }

        return this.jobCredential;
    }

    public DelegatedCredential getStagingCredential()
        throws                              ResourceException,
                                            DelegationException,
                                            GeneralSecurityException,
                                            IOException
    {
        if (logger.isDebugEnabled()) {
            this.logger.debug("getStagingCredential() called");
        }
        if (this.stagingCredential == null) {
            JobDescriptionType jobDescription = getJobDescription();

            String localUserId
                = (String) this.getResourcePropertySet().get(
                    ManagedJobConstants.RP_LOCAL_USER_ID).get(0);

            EndpointReferenceType credentialEndpoint
                = jobDescription.getStagingCredentialEndpoint();
            if (credentialEndpoint == null)
            {
                return null;
            }
            if (logger.isDebugEnabled()) {
                this.logger.debug("staging credential endpoint:\n"
                            + credentialEndpoint);
            }

            String userSubject
                = (String) getResourceProperty(
                    ManagedJobConstants.RP_USER_SUBJECT);

            this.stagingCredential
                = DelegatedCredential.getDelegatedCredential(
                    credentialEndpoint,
                    localUserId,
                    userSubject);

            if (logger.isDebugEnabled()) {
                this.logger.debug("Delegated Staging Credential Object: "
                            +this.stagingCredential.toString());
                try {
                    this.logger.debug(
                        "Using proxy with subject "
                        +this.stagingCredential.getCredential().getIdentity()
                        +"to contact the staging service");
                } catch (Exception e) {
                    String errorMessage = i18n.getMessage(
                        Resources.DELEGATED_PROXY_SUBJECT_ERROR);
                    this.logger.error(errorMessage);
                }
            }
        }

        return this.stagingCredential;
    }

    public ManagedJobResourceDataType getResourceData() {
        return this.resourceData;
    }

    public void setResourceData(ManagedJobResourceDataType resourceData) {
        this.resourceData = resourceData;
    }

    public StateEnumeration getState()
    {
        ResourceProperty stateRP = this.getResourcePropertySet().get(
            ManagedJobConstants.RP_STATE);
        return (StateEnumeration) stateRP.get(0);
    }

    public void setState(StateEnumeration state)
        throws                              Exception
    {
        ResourceProperty stateRP = this.getResourcePropertySet().get(
            ManagedJobConstants.RP_STATE);
        StateEnumeration currentState = (StateEnumeration) stateRP.get(0);

        //set the new state only if the current state isn't terminal
        if (   !currentState.equals(StateEnumeration.Done)
            && !currentState.equals(StateEnumeration.Failed))
        {
            if (logger.isDebugEnabled()) {
                this.logger.debug(
                    "Setting new state RP value " + state.getValue());
            }

            stateRP.set(0, state);
        }
        else
        {
            if (logger.isDebugEnabled()) {
                this.logger.debug(
                    "not setting new state RP value " + state.getValue()
                    + ", since the current state is a terminating state");
            }
        }
    }
}
