/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.multi;

import java.rmi.RemoteException;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.generated.FaultResourcePropertyType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.ManagedMultiJobResourceDataType;
import org.globus.exec.generated.ManagedMultiJobResourcePropertiesType;
import org.globus.exec.generated.ManagedMultiJobResourceStateType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.generated.TopicListenerDescriptionType;
import org.globus.exec.service.job.PersistentManagedJobResource;
import org.globus.exec.service.utils.PersistenceHelper;
import org.globus.exec.utils.ManagedMultiJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.util.I18n;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.impl.PersistentResourceProperty;

import org.oasis.wsrf.faults.BaseFaultType;


public class PersistentManagedMultiJobResource
    extends     ManagedMultiJobResource
    implements  PersistentManagedJobResource
{
    private Object dirtyCond = new Object();

    public void initializeWholeState(
        Object                                  wholeResourceState,
        QName                                   resourceElementQName,
        Object                                  key)
        throws                                  ResourceException
    {
        //this is just to reload the internal rendezvous state data.
        //We do not need/want to do this is the resource is not persistent.

        super.initializeWholeState(
            wholeResourceState, resourceElementQName, key);

        ManagedMultiJobResourceStateType wholeJobState =
            (ManagedMultiJobResourceStateType)wholeResourceState;

        this.setInternalRendezvousState(wholeJobState.getResourceData().
                                        getInternalRendezvousData());
    }

    public void run()
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering run()");
        }

        super.run();

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving run()");
        }
    }

    public void processStateNotification(
        StateEnumeration                    state,
        boolean                             holding,
        FaultResourcePropertyType           fault)
        throws                              Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processStateNotification()");
        }

        super.processStateNotification(state, holding, fault);

        synchronized (dirtyCond)
        {

            setDirty(true);
            store();
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving processStateNotification()");
        }
    }

    public void setFault(FaultType fault)
    {
        super.setFault(fault);

        synchronized (dirtyCond)
        {
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, fault);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    public void setState(StateEnumeration state)
        throws                              Exception
    {
        super.setState(state);

        synchronized (dirtyCond)
        {
            this.setDirty(true); //not needed...?
            this.store();
        }
    }

    public void setResourceProperty(
        QName                               name,
        Object                              value)
    {
        super.setResourceProperty(name, value);

        synchronized (dirtyCond)
        {
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, name);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    public void setResourceDatum(
        String                              name,
        Object                              value)
    {
        setResourceDatumWithoutStoring(name, value);
        try {
            store();
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.POST_UPDATE_PERSISTENCE_ERROR, name);
            throw new RuntimeException(errorMessage, e);
        }
    }

    public void setResourceDatumWithoutStoring(
        String                              name,
        Object                              value)
    {
        super.setResourceDatum(name, value);

        synchronized (dirtyCond)
        {
            setDirty(true);
        }
    }



    public Class getWholeStateType() {
        return ManagedMultiJobResourceStateType.class;
    }

    public QName getWholeStateQName() {
        return WHOLE_STATE_ELEMENT;
    }

    private static final QName WHOLE_STATE_ELEMENT = new QName(
                ManagedMultiJobConstants.MULTI_JOB_STATE_NS,
                "ManagedMultiJobResourceState");

    public QName getResourcePropertiesQName() {
        return ManagedMultiJobConstants.RESOURCE_PROPERTY_SET;
    }

    public Object getWholeState() {
        ManagedMultiJobResourceStateType resourceState =
            new ManagedMultiJobResourceStateType();
        resourceState.setResourceProperties(
            (ManagedMultiJobResourcePropertiesType)this.getResourceBean());
        resourceState.setResourceData(
            (ManagedMultiJobResourceDataType) this.getResourceData());

        //add internal rendezvous implementation state before persisting.
        //we could put this in an overriden getResourceData(), but it's not
        //needed for every call (well, overhead might be small anyway...).
        //store() might be good spot to put it, but I'd like to keep store()
        //simple...
        getResourceData().setInternalRendezvousData(
            this.getInternalRendezvousState());

        return resourceState;
    }

    /**
     * <p>
     * If the resource was successfully stored, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     * @throws ResourceException if the resource could not be stored
     */
    public synchronized void store()
        throws                              ResourceException
    {
        getPersistenceHelper().store();
    }

    /**
     * <p>
     * If the resource was successfully loaded, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     *
     */
    public synchronized void load(ResourceKey key) throws ResourceException {
        getPersistenceHelper().load(key);
    }

    public void remove()
        throws                              ResourceException
    {
        logger.trace("calling super class remove()");
        ResourceException exception = null;
        try
        {
            super.remove();
        }
        catch (ResourceException re)
        {
            exception = re;
        }

        logger.trace("removing persistence data");
        getPersistenceHelper().remove();

        if (exception != null)
        {
            throw exception;
        }
    }

    private synchronized PersistenceHelper getPersistenceHelper()
        throws ResourceException{
        if (this.persistenceHelper == null) {
            try {
                this.persistenceHelper = new PersistenceHelper(this);
            }
            catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.PERSISTENCE_HELPER_CREATION_ERROR);
                logger.error(errorMessage, e);
                throw new ResourceException(errorMessage, e);
            }
        }
        return this.persistenceHelper;
    }

    private PersistenceHelper persistenceHelper;


    protected ResourceProperty createNewResourceProperty(
        QName rpQName,
        Object resourceBean)
        throws Exception {
        ResourceProperty prop =
            super.createNewResourceProperty(rpQName, resourceBean);

        return new PersistentResourceProperty(prop, this);
    }

    /**
     * Sets the dirty flag on this persistent object. The resource will not be
     * persisted unless it the firty flag is true.
     *
     * @param changed boolean To flag the resource as changed i.e. "dirty" and
     * have it persisted by the next invocation of {@link #store() store}
     */
    public void setDirty(boolean changed)
    {
        this.dirty = changed;
    }

    public boolean getDirty() {
        return this.dirty;
    }

    /**
     * The dirty flag is set to true at instanciation time, so that the
     * initial creator of the resource just needs to call {@link #store() store}
     * in order to actually persist the new resource for the first time.
     */
    private boolean dirty = true; //in most cases,
    //creation will occur after a persisted version of the resource exists
    //already in the persistence store.



    private static Log logger =
        LogFactory.getLog(PersistentManagedMultiJobResource.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
    private static I18n coreI18n = I18n.getI18n(Resources.class.getName());

}
