/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.security.GeneralSecurityException;
import java.util.Hashtable;
import java.util.Iterator;

import javax.security.auth.Subject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.globus.delegation.DelegationException;
import org.globus.delegation.DelegationListener;
import org.globus.delegation.DelegationUtil;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.utils.Resources;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.jaas.GlobusPrincipal;
import org.globus.util.I18n;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityConfig;
import org.globus.wsrf.utils.PerformanceLog;

public class DelegatedCredential implements DelegationListener
{
    static Log logger = LogFactory.getLog(DelegatedCredential.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static PerformanceLog performanceLogger = new PerformanceLog(
        DelegatedCredential.class.getName() + ".performance");


    //the real meat of this delegated credential
    GlobusCredential credential = null;

    String localUserId = null;
    File userProxyFile = null;
    private String delegationKey = null;
    private String userSubject;
    private boolean initialized = false;
    private String listenerId = null;

    private static Hashtable INSTANCES;
    static
    {
        INSTANCES = new Hashtable();
    }

    /**
     *  Fetch a delegated credential from the delegation service.
     */
    public static DelegatedCredential getDelegatedCredential(
        EndpointReferenceType               delegationEndpoint,
        String                              localUserId,
        String                              userSubject)
        throws                              DelegationException,
                                            GeneralSecurityException,
                                            IOException
    {
        return getDelegatedCredential(
            delegationEndpoint,
            localUserId,
            userSubject,
            null);
    }

    /**
     *  Fetch a delegated credential from the delegation service and
     *  generate a user proxy file from it.
     */
    public static DelegatedCredential getDelegatedCredential(
        EndpointReferenceType               delegationEndpoint,
        String                              localUserId,
        String                              userSubject,
        File                                userProxyDirectory)
        throws                              DelegationException,
                                            GeneralSecurityException,
                                            IOException
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("Entering getDelegatedCredential()");
        }

        if (logger.isTraceEnabled())
        {
            logger.debug("getting instance table for user " + localUserId);
        }
        Hashtable userInstances = null;
        synchronized (INSTANCES)
        {
            userInstances = (Hashtable) INSTANCES.get(localUserId);
            if (userInstances == null)
            {
                userInstances = new Hashtable();
                INSTANCES.put(localUserId, userInstances);
            }
        }

        if (logger.isTraceEnabled())
        {
            logger.debug("checking for existing credential listener");
        }
        DelegatedCredential cred = null;
        synchronized (userInstances)
        {
            String delegationKey = getDelegationKey(delegationEndpoint);
            if (userInstances.containsKey(delegationKey))
            {
                DelegatedCredential savedCred
                    = (DelegatedCredential) userInstances.get(delegationKey);

                if (   (userProxyDirectory != null)
                    && (savedCred.userProxyFile == null))
                {
                    //already initialized except for the user proxy
                    savedCred.userProxyFile = new File(
                        userProxyDirectory,
                        "gram_job_proxy_" + savedCred.delegationKey);
                    if (logger.isDebugEnabled())
                    {
                        logger.debug("userProxFile: "
                                    + savedCred.userProxyFile.toString());
                    }

                    new UserProxyCreator(savedCred).run();
                }

                return savedCred;
            }

            if (logger.isDebugEnabled()) {
                logger.debug("delegated credential endpoint (BEFORE): "
                            + delegationEndpoint);
            }

            if (logger.isTraceEnabled())
            {
                logger.trace("new DelegatedCredential()");
            }
            performanceLogger.start();
            cred = new DelegatedCredential(
                delegationEndpoint,
                localUserId,
                userSubject,
                userProxyDirectory);
            performanceLogger.stop("new DelegatedCredential()");

            if (logger.isDebugEnabled()) {
                logger.debug("delegated credential endpoint (AFTER): "
                            + delegationEndpoint);
            }

            Subject subject = new Subject();
            subject.getPrincipals().add(new GlobusPrincipal(userSubject));

            performanceLogger.start();
            try
            {
                DelegationUtil.registerDelegationListener(
                    delegationEndpoint,
                    cred,
                    subject);
            }
            catch (DelegationException de)
            {
                cred.credentialDeleted();
                throw de;
            }

            performanceLogger.stop(
                "DelegationUtil.registerDelegationListener()");

            performanceLogger.start();
            cred.initialize();
            performanceLogger.stop("cred.initialize()");
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving getDelegatedCredential()");
        }

        return cred;
    }

    /** Creates a new delegated credential by fetching the actual
     *  credential from the Delegation Service, and stores an
     *  additional user-owned proxy in the specified location.
     */
    private DelegatedCredential(
        EndpointReferenceType               delegationEndpoint,
        String                              localUserId,
        String                              userSubject,
        File                                userProxyDirectory)
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("Entering DelegatedCredential()");
        }

        this.localUserId = localUserId;
        this.userSubject = userSubject;

        this.delegationKey
            = DelegatedCredential.getDelegationKey(delegationEndpoint);

        if (userProxyDirectory != null)
        {
            this.userProxyFile = new File(
                userProxyDirectory,
                "gram_job_proxy_" + this.delegationKey);
            if (logger.isDebugEnabled())
            {
                logger.debug("userProxFile: " + this.userProxyFile.toString());
            }
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("saving this instance to instance table");
        }

        Hashtable userInstances = null;

        synchronized (INSTANCES)
        {
            userInstances = (Hashtable) INSTANCES.get(localUserId);
        }

        synchronized (userInstances)
        {
            userInstances.put(this.delegationKey, this);
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving DelegatedCredential()");
        }
    }

    private void initialize()
        throws                              DelegationException,
                                            GeneralSecurityException,
                                            IOException
    {
        try {
            synchronized (this) {
                if (logger.isDebugEnabled()) {
                    logger.debug("waiting to receive initial credential...");
                }
                while (!this.initialized) {
                    wait();
                }
                if (logger.isDebugEnabled()) {
                    logger.debug("done waiting...");
                }
            }
        } catch (Exception e) {
            String errorMessage
                = i18n.getMessage(Resources.INTERRUPTED_CRED_WAIT);
            logger.error(errorMessage, e);
        }
    }

    /** The credential refresh listener callback.
     */
    public void setCredential(
        GlobusCredential                    credential)
    {
        if (logger.isTraceEnabled()) {
            logger.trace("Entering setCredential()");
        }

        //if this method was called because the DS resource was recovered,
        //then re-add the instance to the instance table
        Hashtable userInstances = null;
        synchronized (INSTANCES)
        {
            userInstances = (Hashtable) INSTANCES.get(localUserId);
            if (userInstances == null)
            {
                userInstances = new Hashtable();
                INSTANCES.put(this.localUserId, userInstances);
            }
        }
        synchronized (userInstances)
        {
            if (userInstances.get(this.delegationKey) == null)
            {
                userInstances.put(this.delegationKey, this);
            }
        }

        this.credential = credential;

        if (logger.isDebugEnabled()) {
            logger.debug("received a new credential:\n"
                        +this.credential);
        }

        //create and save a user proxy to disk based on this credential
        if (this.userProxyFile != null)
        {

            //create in a separate thread
            //new Thread(new UserProxyCreator(this)).start();

            //create in the current thread
            //performanceLogger.start();
            if (logger.isTraceEnabled()) {
                logger.trace("at run()");
            }
            new UserProxyCreator(this).run();
            //performanceLogger.stop("createUserProxy()");
        } else
        {
            if (logger.isTraceEnabled()) {
                logger.trace("at setInitialized()");
            }
            setInitialized();
        }

        if (logger.isTraceEnabled()) {
            logger.trace("Leaving setCredential()");
        }
    }

    synchronized void setInitialized()
    {
        if (!this.initialized) {
            this.initialized = true;
            notifyAll();
        }
    }

    /** No use in having a delegated credential if you can't use it!
     */
    public GlobusCredential getCredential()
    {
        try {
            if (logger.isDebugEnabled()) {
                logger.debug("waiting to receive initial credential...");
            }
            synchronized (this) {
                while (!this.initialized) {
                    wait();
                }
            }
            if (logger.isDebugEnabled()) {
                logger.debug("done waiting...");
            }
        } catch (Exception e) {
            String errorMessage
                = i18n.getMessage(Resources.INTERRUPTED_CRED_WAIT);
            logger.error(errorMessage, e);
        }

        return this.credential;
    }

    public void removeUserProxy()
        throws                              GeneralSecurityException,
                                            IOException
    {
        if ((this.userProxyFile == null) || !this.userProxyFile.exists()) {
            if (logger.isDebugEnabled()) {
                logger.debug("Ignoring job user proxy removal request.");
            }
            return;
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Removing job user proxy.");
        }

        if (!System.getProperty("user.name").equals(this.localUserId))
        {
            //Get shell tool paths
            GlobusShToolsProperties toolPathProps
                = GlobusShToolsProperties.getInstance();

            //sudo <username> globus-gram-local-proxy-tool \
            //      -delete <user proxy file>

            String sudo = toolPathProps.getProperty("GLOBUS_SH_SUDO");

            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String gridMapAndExecute
                = new File(factoryHomeConfiguration.getGlobusLocation()
                + "/libexec/globus-gridmap-and-execute").toString();
            String proxyTool
                = new File(factoryHomeConfiguration.getGlobusLocation()
                + "/libexec/globus-gram-local-proxy-tool").toString();
            String gridMapFile = null;
            try {
                gridMapFile
                    = ContainerSecurityConfig.getConfig()
                    .getSecurityDescriptor().getGridMapFile();
            } catch(Exception e)
            {
                throw new RuntimeException(e);
            }

            String[] command = null;
            if (gridMapFile != null)
            {
                command = new String[] {
                    sudo,
                    "-u",
                    this.localUserId,
                    "-S",
                    gridMapAndExecute,
                    "-g",
                    gridMapFile,
                    proxyTool,
                    factoryHomeConfiguration.getGlobusLocation(),
                    "-delete",
                    this.userProxyFile.getPath()
                };
            } else
            {
                command = new String[] {
                    sudo,
                    "-u",
                    this.localUserId,
                    "-S",
                    gridMapAndExecute,
                    proxyTool,
                    factoryHomeConfiguration.getGlobusLocation(),
                    "-delete",
                    this.userProxyFile.getPath()
                };
            }
            if (logger.isDebugEnabled()) {
                StringBuffer commandBuffer = new StringBuffer(command.length*2);
                for (int index=0; index<command.length; index++) {
                    if (index > 0) {
                        commandBuffer.append(" ");
                    }
                    commandBuffer.append(command[index]);
                }
                logger.debug("Executing command:\n" + commandBuffer.toString());
            }

            Process proc = Runtime.getRuntime().exec(command);

            proc.getOutputStream().close();

            //handle possible sudo password prompt
            BufferedReader stderr = new BufferedReader(
                new InputStreamReader(proc.getErrorStream()));
            String stderrLine = stderr.readLine();
            while (stderrLine != null)
            {
                if (stderrLine.trim().indexOf("Password:") == 0)
                {
                    String errorMessage = i18n.getMessage(
                        Resources.MISCONFIGURED_SUDO,
                        new String[] { "globus-gram-local-proxy-tool",
                                       this.localUserId });
                    throw new RuntimeException(errorMessage);
                }
            }
            stderr.close();

            proc.getInputStream().close();

        } else {
            this.userProxyFile.delete();
        }
    }

    public void credentialDeleted()
    {
        if (logger.isTraceEnabled())
        {
            logger.trace("entering credentialDeleted()");
        }

        Hashtable userInstances = null;

        synchronized (INSTANCES)
        {
            userInstances = (Hashtable) INSTANCES.get(localUserId);
        }

        synchronized (userInstances)
        {
            DelegatedCredential credObj = (DelegatedCredential)
                userInstances.remove(this.delegationKey);
            if (credObj != null)
            {
                synchronized (credObj)
                {
                    credObj.notifyAll();
                }
                logger.debug("removed DelegatedCredential object");
            }
        }

        //remove user proxy file if it exists
        try {
            removeUserProxy();
        } catch (Exception e)
        {
            logger.warn("unable to remove user proxy", e);
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("leaving credentialDeleted()");
        }
    }

    /* Begin serialization methods */
    public DelegatedCredential() { }

    /* The getCredential() and setCredential() methods are
     * already implemented for reasons other than serialization.
     */
    public String getLocalUserId()
    {
        return this.localUserId;
    }
    public void setLocalUserId(String localUserId)
    {
        this.localUserId = localUserId;
    }
    public String getUserSubject()
    {
        return this.userSubject;
    }
    public void setUserSubject(String userSubject)
    {
        this.userSubject = userSubject;
    }
    public File getUserProxyFile()
    {
        return this.userProxyFile;
    }
    public void setUserProxyFile(File userProxyFile)
    {
        this.userProxyFile = userProxyFile;
    }
    public boolean getInitialized()
    {
        return this.initialized;
    }
    public void setInitialized(boolean initialized)
    {
        this.initialized = initialized;
    }
    public String getDelegationKey()
    {
        return this.delegationKey;
    }
    public void setDelegationKey(String delegationKey)
    {
        this.delegationKey = delegationKey;
    }
    /* End serialization methods */

    public String getId()
    {
        return this.listenerId;
    }
    public void setId(String listenerId)
    {
        this.listenerId = listenerId;
    }

    private static String getDelegationKey(
        EndpointReferenceType               credentialEndpoint)
    {
        MessageElement[] refProps
            = credentialEndpoint.getProperties().get_any();
        String delegationKey = null;
        try {
            delegationKey = ((String) ObjectDeserializer.
                toObject(refProps[0], String.class)).trim();
        } catch (Exception e)
        {
            throw new RuntimeException(
                "Unable to deserialize DelegationKey", e);
        }

        if (logger.isDebugEnabled())
        {
            logger.debug("Pulled out DelegationKey: " + delegationKey);
        }
        return delegationKey;
    }
}
