/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import java.io.IOException;

import javax.xml.namespace.QName;

import org.globus.util.I18n;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.utils.XmlPersistenceHelper;
import org.globus.wsrf.utils.PerformanceLog;

import org.globus.exec.service.job.PersistentManagedJobResource;
import org.globus.exec.service.job.ManagedJobResourceImpl;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.generated.TopicListenerDescriptionType;

public class PersistenceHelper extends XmlPersistenceHelper
{
    private static Log logger = LogFactory.getLog(
        PersistenceHelper.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static PerformanceLog performanceLogger = new PerformanceLog(
        PersistenceHelper.class.getName() + ".performance");

    private static final boolean TESTING = false; //conditional compilation

    private PersistentManagedJobResource jobResource;

    /**
     * <b>Postcondition<n/> this.getResource().equals(resource)
     * @param resource PersistentManagedJobResource
     * @param wholeStateSerializableType Class
     * @param wholeStateXMLElementQName QName this is only used for storing
     * @throws IOException
     */
    public PersistenceHelper(PersistentManagedJobResource resource)
        throws IOException {
        super(resource.getWholeStateType());
        this.jobResource = resource;
    }

    public PersistentManagedJobResource getResource() {
        return this.jobResource;
    }

    /**
     * <p>
     * If the resource was successfully stored, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     * @see #getKeyAsFile(Object) getKeyAsFile()
     * @see PersistentResource
     * @throws ResourceException if the resource could not be stored
     */
    public synchronized void store() throws ResourceException {

        if (!this.jobResource.getDirty()) {
            logger.debug("Dirty flag not set: resource will not be stored");
            return;
        }

        if (logger.isDebugEnabled()) {
            logger.debug("storing resource " + this.jobResource.getID()
                        +" with state " + this.jobResource.getState());
        }


        //set the internal state topic listeners in order to persist
        //subscriptions. we don't need to do this anywhere except before
        //persisting (it's only for persistence).
        try {
            ManagedJobResourceImpl.setResourceDatum(
                this.jobResource,
                ManagedJobConstants.RD_TOPIC_LISTENER,
                SubscriptionPersistenceUtils.getSerializableListeners(
                    this.jobResource.getTopicList()));
        }
        catch (Exception e) {
            String errorMessage =
                i18n.getMessage(Resources.TOPIC_SERIALIZ_ERROR);
            logger.error(errorMessage, e);
            throw new RuntimeException(errorMessage, e);
        }

        /*
        //add rendezvous internal data to whole state of resource:
        this.jobResource.setResourceDatum(
                "internalRendezvousData",
                this.jobResource.getInternalRendezvousState());*/

        this.store(
            this.jobResource.getID(),
            this.jobResource.getWholeState(),
            this.jobResource.getWholeStateQName());

        if (TESTING) {
            this.load( (ResourceKey) this.jobResource.getID());
        }

        this.jobResource.setDirty(false);
        if (logger.isDebugEnabled()) {
            logger.debug("done storing resource " + this.jobResource.getID());
        }

    }

    /**
     * <p>
     * If the resource was successfully loaded, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     *
     */
    public synchronized void load(ResourceKey key) throws ResourceException {
        performanceLogger.start();

        if (logger.isDebugEnabled()) {
            logger.debug("loading resource " + this.jobResource +
                         " of key " + key);
        }

        //load resource properties and internal state

        Object loadedResourceStateBean = super.load(key);

        this.jobResource.initializeWholeState(
            loadedResourceStateBean,
            this.jobResource.getResourcePropertiesQName(),
            key);

        //read persisted subscriptions from the deserialized topic listeners
        //in the internal state and register listeners with topics.
        try {
            if (logger.isDebugEnabled()) {
                logger.debug("reconstructing topic listeners based on " +
                             "deserialized persisted subscriptions");
            }

            TopicListenerDescriptionType[] serializableListeners =
                (TopicListenerDescriptionType[])this.jobResource.
                getResourceDatum(ManagedJobConstants.RD_TOPIC_LISTENER);

            SubscriptionPersistenceUtils.addSubscriptionListenersToTopics(
                serializableListeners,
                this.jobResource.getTopicList());
        }
        catch (Exception e) {
            String errorMessage =
                i18n.getMessage(Resources.TOPIC_REGISTRATION_ERROR);
            logger.error(errorMessage, e);
            throw new RuntimeException(errorMessage, e);
        }

        this.jobResource.setDirty(false);
        if (logger.isDebugEnabled()) {
            logger.debug("done loading  resource " + this.jobResource.getID()
                        +" with state " + this.jobResource.getState());
        }
        performanceLogger.stop("load(ResourceKey)");
    }

    public void remove() throws ResourceException {
        if (logger.isDebugEnabled()) {
            logger.debug("Removing XML resource state file");
            logger.debug("resource class: "
                        + this.jobResource.getClass().getName());
            logger.debug("resource id: "
                        + this.jobResource.getID().toString());
        }
        if (!TESTING) {
            super.remove(this.jobResource.getID());
        }
    }
}
