/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Vector;

import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.common.ChainedIOException;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.utils.Resources;
import org.globus.gsi.CertUtil;
import org.globus.gsi.GSIConstants;
import org.globus.gsi.bc.BouncyCastleCertProcessingFactory;
import org.globus.util.Base64;
import org.globus.util.I18n;
import org.globus.util.PEMUtils;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityConfig;
import org.globus.wsrf.utils.PerformanceLog;

public class UserProxyCreator implements Runnable
{
    static Log logger = LogFactory.getLog(UserProxyCreator.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static PerformanceLog performanceLogger = new PerformanceLog(
        UserProxyCreator.class.getName() + ".performance");

    private DelegatedCredential credential = null;

    public UserProxyCreator(
        DelegatedCredential                 credential)
    {
        this.credential = credential;
    }

    /** Create a new credential with a new key pair signed by the
     *  delegation private key and including the delegated cert chain.
     *  The credential is used for outbound transactions from processes
     *  running under a user account.
     */
    public void run()
    {
        InputStream stdout = null;
        OutputStream proxyFileOutputStream = null;

        try
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Creating job user proxy.");
            }

            X509Certificate userProxy = null;

            //Get shell tool paths
            performanceLogger.start();
            GlobusShToolsProperties toolPathProps
                = GlobusShToolsProperties.getInstance();

            Process proc = null;
            //sudo <username> globus-gram-local-proxy-tool \
            //      -write <user proxy file>

            String sudo = toolPathProps.getProperty("GLOBUS_SH_SUDO");
            if ((sudo == null) || (sudo.length() == 0))
            {
                throw new IOException(i18n.getMessage(Resources.MISSING_SUDO));
            }

            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String proxyTool
                = new File(factoryHomeConfiguration.getGlobusLocation()
                  + "/libexec/globus-gram-local-proxy-tool").toString();

            String gridMapAndExecute
                = new File(factoryHomeConfiguration
                .getGlobusLocation()
                + "/libexec/globus-gridmap-and-execute").toString();
            String gridMapFile
                = ContainerSecurityConfig.getConfig()
                .getSecurityDescriptor().getGridMapFile();
            boolean authzGridmap
                = AuthorizationHelper.isAuthorizationGridmap();

            Vector commandVector = new Vector();
            if (!System.getProperty("user.name").equals(
                this.credential.localUserId))
            {
                commandVector.add(sudo);
                commandVector.add("-u");
                commandVector.add(this.credential.localUserId);
                commandVector.add("-S");
                if (authzGridmap)
                {
                    commandVector.add(gridMapAndExecute);
                    if (gridMapFile != null)
                    {
                        commandVector.add("-g");
                        commandVector.add(gridMapFile);
                    }
                }
            }
            commandVector.add(proxyTool);
            commandVector.add(
                factoryHomeConfiguration.getGlobusLocation());
            commandVector.add("-write");
            commandVector.add(
                this.credential.userProxyFile.getPath());

            if (logger.isDebugEnabled())
            {
                StringBuffer commandBuffer
                    = new StringBuffer(commandVector.size()*2);
                for (int index=0; index<commandVector.size(); index++)
                {
                    if (index > 0)
                    {
                        commandBuffer.append(" ");
                    }
                    commandBuffer.append(commandVector.get(index));
                }
                if (logger.isDebugEnabled())
                {
                    logger.debug("Executing command:\n"
                                + commandBuffer.toString());
                }
            }

            proc = Runtime.getRuntime().exec(
                (String[]) commandVector.toArray(new String[0]));

            stdout = proc.getInputStream();


            int proxyType;

            if (CertUtil.isLimitedProxy(
                this.credential.credential.getProxyType()))
            {
                //create a new limited proxy credential
                proxyType = GSIConstants.DELEGATION_LIMITED;
            } else {
                //create a new full proxy credential
                proxyType = GSIConstants.DELEGATION_FULL;
            }

            InputStream derRequest = getDERRequest(stdout);
            if (derRequest == null)
            {
                //handle possible sudo password prompt
                BufferedReader stderr = new BufferedReader(
                    new InputStreamReader(proc.getErrorStream()));
                String stderrLine = stderr.readLine();
                while (stderrLine != null)
                {
                    if (stderrLine.trim().indexOf("Password:") == 0)
                    {
                        String errorMessage = i18n.getMessage(
                            Resources.MISCONFIGURED_SUDO,
                            new String[] { "globus-gram-local-proxy-tool",
                                           this.credential.localUserId });
                        throw new RuntimeException(errorMessage);
                    }
                }
                stderr.close();
            }

            if (logger.isTraceEnabled())
            {
                logger.trace("at createCertificate()");
            }
            performanceLogger.start();
            userProxy = BouncyCastleCertProcessingFactory.getDefault()
                .createCertificate(
                    derRequest,
                    this.credential.credential.getCertificateChain()[0],
                    this.credential.credential.getPrivateKey(),
                    (int) this.credential.credential.getTimeLeft(),
                    proxyType);
            performanceLogger.stop("createCertificate()");

            //setup an output stream to stdin of tee (running as the user)
            proxyFileOutputStream =
                new BufferedOutputStream(proc.getOutputStream());
            if (logger.isTraceEnabled())
            {
                logger.trace("at writeProxy()");
            }
            writeProxy(userProxy,
                       this.credential.credential.getCertificateChain(),
                       proxyFileOutputStream);
            proxyFileOutputStream.close();
            proxyFileOutputStream = null;
            
            int rc = 0;
            try {
                rc = proc.waitFor();
            } catch (InterruptedException ie) { }
            if (rc != 0) {
                this.credential.removeUserProxy();

                throw new IOException(
                    I18n.getI18n(Resources.class.getName()).
                    getMessage(
                        "changeOwnerFailed",
                        new String[] {
                            this.credential.localUserId,
                            this.credential.userProxyFile.getPath()
                        }));
            }
            performanceLogger.stop("write user proxy");
        }
        catch (Exception e)
        {
            throw new RuntimeException(
                i18n.getMessage(Resources.PROXY_FILE_CREATION_ERROR), e);
        } finally {
            try {
                if (stdout != null) {
                    stdout.close();
                }
                if (proxyFileOutputStream != null) {
                    proxyFileOutputStream.close();
                }
            } catch (IOException e) {
            }
        }

        this.credential.setInitialized();
    }

    private static final void writePEM(OutputStream out,
                                       X509Certificate cert)
        throws CertificateEncodingException, IOException {
        PEMUtils.writeBase64(out,
                             "-----BEGIN CERTIFICATE-----",
                             Base64.encode(cert.getEncoded()),
                             "-----END CERTIFICATE-----");
    }

    private static void writeProxy(X509Certificate proxy,
                                   X509Certificate[] chain,
                                   OutputStream out)
        throws IOException {
        try {
            writePEM(out, proxy);

            for (int i = 0; i < chain.length; i++) {
                // this will skip the self-signed certificates
                if (chain[i].getSubjectDN().equals(chain[i].getIssuerDN()))
                {
                    continue;
                }
                writePEM(out, chain[i]);
            }
        } catch (CertificateEncodingException e) {
            throw new ChainedIOException(e.getMessage(), e);
        }
        out.flush();
    }

    private static InputStream getDERRequest(InputStream in) throws IOException
    {
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        String line = null;
        StringBuffer buff = new StringBuffer();
        boolean isReq = false;
        while ((line = reader.readLine()) != null) {
            if (line.indexOf("BEGIN CERTIFICATE REQUEST") != -1) {
                isReq = true;
            } else if (isReq && line.indexOf("END CERTIFICATE REQUEST") != -1) {
                byte [] data = Base64.decode(buff.toString().getBytes());
                return new ByteArrayInputStream(data);
            } else if (isReq) {
                buff.append(line);
            }
        }
        return null;
    }
}
